# /* Copyright (C) 2001
#  * Housemarque Oy
#  * http://www.housemarque.com
#  *
#  * Distributed under the Boost Software License, Version 1.0. (See
#  * accompanying file LICENSE_1_0.txt or copy at
#  * http://www.boost.org/LICENSE_1_0.txt)
#  */
#
# /* Revised by Paul Mensonides (2002) */
# /* Revised by Edward Diener (2020) */
#
# /* See http://www.boost.org for most recent version. */
#
# ifndef BOOST_PREPROCESSOR_LOGICAL_BOOL_1024_HPP
# define BOOST_PREPROCESSOR_LOGICAL_BOOL_1024_HPP
#
# define BOOST_PP_BOOL_513 1
# define BOOST_PP_BOOL_514 1
# define BOOST_PP_BOOL_515 1
# define BOOST_PP_BOOL_516 1
# define BOOST_PP_BOOL_517 1
# define BOOST_PP_BOOL_518 1
# define BOOST_PP_BOOL_519 1
# define BOOST_PP_BOOL_520 1
# define BOOST_PP_BOOL_521 1
# define BOOST_PP_BOOL_522 1
# define BOOST_PP_BOOL_523 1
# define BOOST_PP_BOOL_524 1
# define BOOST_PP_BOOL_525 1
# define BOOST_PP_BOOL_526 1
# define BOOST_PP_BOOL_527 1
# define BOOST_PP_BOOL_528 1
# define BOOST_PP_BOOL_529 1
# define BOOST_PP_BOOL_530 1
# define BOOST_PP_BOOL_531 1
# define BOOST_PP_BOOL_532 1
# define BOOST_PP_BOOL_533 1
# define BOOST_PP_BOOL_534 1
# define BOOST_PP_BOOL_535 1
# define BOOST_PP_BOOL_536 1
# define BOOST_PP_BOOL_537 1
# define BOOST_PP_BOOL_538 1
# define BOOST_PP_BOOL_539 1
# define BOOST_PP_BOOL_540 1
# define BOOST_PP_BOOL_541 1
# define BOOST_PP_BOOL_542 1
# define BOOST_PP_BOOL_543 1
# define BOOST_PP_BOOL_544 1
# define BOOST_PP_BOOL_545 1
# define BOOST_PP_BOOL_546 1
# define BOOST_PP_BOOL_547 1
# define BOOST_PP_BOOL_548 1
# define BOOST_PP_BOOL_549 1
# define BOOST_PP_BOOL_550 1
# define BOOST_PP_BOOL_551 1
# define BOOST_PP_BOOL_552 1
# define BOOST_PP_BOOL_553 1
# define BOOST_PP_BOOL_554 1
# define BOOST_PP_BOOL_555 1
# define BOOST_PP_BOOL_556 1
# define BOOST_PP_BOOL_557 1
# define BOOST_PP_BOOL_558 1
# define BOOST_PP_BOOL_559 1
# define BOOST_PP_BOOL_560 1
# define BOOST_PP_BOOL_561 1
# define BOOST_PP_BOOL_562 1
# define BOOST_PP_BOOL_563 1
# define BOOST_PP_BOOL_564 1
# define BOOST_PP_BOOL_565 1
# define BOOST_PP_BOOL_566 1
# define BOOST_PP_BOOL_567 1
# define BOOST_PP_BOOL_568 1
# define BOOST_PP_BOOL_569 1
# define BOOST_PP_BOOL_570 1
# define BOOST_PP_BOOL_571 1
# define BOOST_PP_BOOL_572 1
# define BOOST_PP_BOOL_573 1
# define BOOST_PP_BOOL_574 1
# define BOOST_PP_BOOL_575 1
# define BOOST_PP_BOOL_576 1
# define BOOST_PP_BOOL_577 1
# define BOOST_PP_BOOL_578 1
# define BOOST_PP_BOOL_579 1
# define BOOST_PP_BOOL_580 1
# define BOOST_PP_BOOL_581 1
# define BOOST_PP_BOOL_582 1
# define BOOST_PP_BOOL_583 1
# define BOOST_PP_BOOL_584 1
# define BOOST_PP_BOOL_585 1
# define BOOST_PP_BOOL_586 1
# define BOOST_PP_BOOL_587 1
# define BOOST_PP_BOOL_588 1
# define BOOST_PP_BOOL_589 1
# define BOOST_PP_BOOL_590 1
# define BOOST_PP_BOOL_591 1
# define BOOST_PP_BOOL_592 1
# define BOOST_PP_BOOL_593 1
# define BOOST_PP_BOOL_594 1
# define BOOST_PP_BOOL_595 1
# define BOOST_PP_BOOL_596 1
# define BOOST_PP_BOOL_597 1
# define BOOST_PP_BOOL_598 1
# define BOOST_PP_BOOL_599 1
# define BOOST_PP_BOOL_600 1
# define BOOST_PP_BOOL_601 1
# define BOOST_PP_BOOL_602 1
# define BOOST_PP_BOOL_603 1
# define BOOST_PP_BOOL_604 1
# define BOOST_PP_BOOL_605 1
# define BOOST_PP_BOOL_606 1
# define BOOST_PP_BOOL_607 1
# define BOOST_PP_BOOL_608 1
# define BOOST_PP_BOOL_609 1
# define BOOST_PP_BOOL_610 1
# define BOOST_PP_BOOL_611 1
# define BOOST_PP_BOOL_612 1
# define BOOST_PP_BOOL_613 1
# define BOOST_PP_BOOL_614 1
# define BOOST_PP_BOOL_615 1
# define BOOST_PP_BOOL_616 1
# define BOOST_PP_BOOL_617 1
# define BOOST_PP_BOOL_618 1
# define BOOST_PP_BOOL_619 1
# define BOOST_PP_BOOL_620 1
# define BOOST_PP_BOOL_621 1
# define BOOST_PP_BOOL_622 1
# define BOOST_PP_BOOL_623 1
# define BOOST_PP_BOOL_624 1
# define BOOST_PP_BOOL_625 1
# define BOOST_PP_BOOL_626 1
# define BOOST_PP_BOOL_627 1
# define BOOST_PP_BOOL_628 1
# define BOOST_PP_BOOL_629 1
# define BOOST_PP_BOOL_630 1
# define BOOST_PP_BOOL_631 1
# define BOOST_PP_BOOL_632 1
# define BOOST_PP_BOOL_633 1
# define BOOST_PP_BOOL_634 1
# define BOOST_PP_BOOL_635 1
# define BOOST_PP_BOOL_636 1
# define BOOST_PP_BOOL_637 1
# define BOOST_PP_BOOL_638 1
# define BOOST_PP_BOOL_639 1
# define BOOST_PP_BOOL_640 1
# define BOOST_PP_BOOL_641 1
# define BOOST_PP_BOOL_642 1
# define BOOST_PP_BOOL_643 1
# define BOOST_PP_BOOL_644 1
# define BOOST_PP_BOOL_645 1
# define BOOST_PP_BOOL_646 1
# define BOOST_PP_BOOL_647 1
# define BOOST_PP_BOOL_648 1
# define BOOST_PP_BOOL_649 1
# define BOOST_PP_BOOL_650 1
# define BOOST_PP_BOOL_651 1
# define BOOST_PP_BOOL_652 1
# define BOOST_PP_BOOL_653 1
# define BOOST_PP_BOOL_654 1
# define BOOST_PP_BOOL_655 1
# define BOOST_PP_BOOL_656 1
# define BOOST_PP_BOOL_657 1
# define BOOST_PP_BOOL_658 1
# define BOOST_PP_BOOL_659 1
# define BOOST_PP_BOOL_660 1
# define BOOST_PP_BOOL_661 1
# define BOOST_PP_BOOL_662 1
# define BOOST_PP_BOOL_663 1
# define BOOST_PP_BOOL_664 1
# define BOOST_PP_BOOL_665 1
# define BOOST_PP_BOOL_666 1
# define BOOST_PP_BOOL_667 1
# define BOOST_PP_BOOL_668 1
# define BOOST_PP_BOOL_669 1
# define BOOST_PP_BOOL_670 1
# define BOOST_PP_BOOL_671 1
# define BOOST_PP_BOOL_672 1
# define BOOST_PP_BOOL_673 1
# define BOOST_PP_BOOL_674 1
# define BOOST_PP_BOOL_675 1
# define BOOST_PP_BOOL_676 1
# define BOOST_PP_BOOL_677 1
# define BOOST_PP_BOOL_678 1
# define BOOST_PP_BOOL_679 1
# define BOOST_PP_BOOL_680 1
# define BOOST_PP_BOOL_681 1
# define BOOST_PP_BOOL_682 1
# define BOOST_PP_BOOL_683 1
# define BOOST_PP_BOOL_684 1
# define BOOST_PP_BOOL_685 1
# define BOOST_PP_BOOL_686 1
# define BOOST_PP_BOOL_687 1
# define BOOST_PP_BOOL_688 1
# define BOOST_PP_BOOL_689 1
# define BOOST_PP_BOOL_690 1
# define BOOST_PP_BOOL_691 1
# define BOOST_PP_BOOL_692 1
# define BOOST_PP_BOOL_693 1
# define BOOST_PP_BOOL_694 1
# define BOOST_PP_BOOL_695 1
# define BOOST_PP_BOOL_696 1
# define BOOST_PP_BOOL_697 1
# define BOOST_PP_BOOL_698 1
# define BOOST_PP_BOOL_699 1
# define BOOST_PP_BOOL_700 1
# define BOOST_PP_BOOL_701 1
# define BOOST_PP_BOOL_702 1
# define BOOST_PP_BOOL_703 1
# define BOOST_PP_BOOL_704 1
# define BOOST_PP_BOOL_705 1
# define BOOST_PP_BOOL_706 1
# define BOOST_PP_BOOL_707 1
# define BOOST_PP_BOOL_708 1
# define BOOST_PP_BOOL_709 1
# define BOOST_PP_BOOL_710 1
# define BOOST_PP_BOOL_711 1
# define BOOST_PP_BOOL_712 1
# define BOOST_PP_BOOL_713 1
# define BOOST_PP_BOOL_714 1
# define BOOST_PP_BOOL_715 1
# define BOOST_PP_BOOL_716 1
# define BOOST_PP_BOOL_717 1
# define BOOST_PP_BOOL_718 1
# define BOOST_PP_BOOL_719 1
# define BOOST_PP_BOOL_720 1
# define BOOST_PP_BOOL_721 1
# define BOOST_PP_BOOL_722 1
# define BOOST_PP_BOOL_723 1
# define BOOST_PP_BOOL_724 1
# define BOOST_PP_BOOL_725 1
# define BOOST_PP_BOOL_726 1
# define BOOST_PP_BOOL_727 1
# define BOOST_PP_BOOL_728 1
# define BOOST_PP_BOOL_729 1
# define BOOST_PP_BOOL_730 1
# define BOOST_PP_BOOL_731 1
# define BOOST_PP_BOOL_732 1
# define BOOST_PP_BOOL_733 1
# define BOOST_PP_BOOL_734 1
# define BOOST_PP_BOOL_735 1
# define BOOST_PP_BOOL_736 1
# define BOOST_PP_BOOL_737 1
# define BOOST_PP_BOOL_738 1
# define BOOST_PP_BOOL_739 1
# define BOOST_PP_BOOL_740 1
# define BOOST_PP_BOOL_741 1
# define BOOST_PP_BOOL_742 1
# define BOOST_PP_BOOL_743 1
# define BOOST_PP_BOOL_744 1
# define BOOST_PP_BOOL_745 1
# define BOOST_PP_BOOL_746 1
# define BOOST_PP_BOOL_747 1
# define BOOST_PP_BOOL_748 1
# define BOOST_PP_BOOL_749 1
# define BOOST_PP_BOOL_750 1
# define BOOST_PP_BOOL_751 1
# define BOOST_PP_BOOL_752 1
# define BOOST_PP_BOOL_753 1
# define BOOST_PP_BOOL_754 1
# define BOOST_PP_BOOL_755 1
# define BOOST_PP_BOOL_756 1
# define BOOST_PP_BOOL_757 1
# define BOOST_PP_BOOL_758 1
# define BOOST_PP_BOOL_759 1
# define BOOST_PP_BOOL_760 1
# define BOOST_PP_BOOL_761 1
# define BOOST_PP_BOOL_762 1
# define BOOST_PP_BOOL_763 1
# define BOOST_PP_BOOL_764 1
# define BOOST_PP_BOOL_765 1
# define BOOST_PP_BOOL_766 1
# define BOOST_PP_BOOL_767 1
# define BOOST_PP_BOOL_768 1
# define BOOST_PP_BOOL_769 1
# define BOOST_PP_BOOL_770 1
# define BOOST_PP_BOOL_771 1
# define BOOST_PP_BOOL_772 1
# define BOOST_PP_BOOL_773 1
# define BOOST_PP_BOOL_774 1
# define BOOST_PP_BOOL_775 1
# define BOOST_PP_BOOL_776 1
# define BOOST_PP_BOOL_777 1
# define BOOST_PP_BOOL_778 1
# define BOOST_PP_BOOL_779 1
# define BOOST_PP_BOOL_780 1
# define BOOST_PP_BOOL_781 1
# define BOOST_PP_BOOL_782 1
# define BOOST_PP_BOOL_783 1
# define BOOST_PP_BOOL_784 1
# define BOOST_PP_BOOL_785 1
# define BOOST_PP_BOOL_786 1
# define BOOST_PP_BOOL_787 1
# define BOOST_PP_BOOL_788 1
# define BOOST_PP_BOOL_789 1
# define BOOST_PP_BOOL_790 1
# define BOOST_PP_BOOL_791 1
# define BOOST_PP_BOOL_792 1
# define BOOST_PP_BOOL_793 1
# define BOOST_PP_BOOL_794 1
# define BOOST_PP_BOOL_795 1
# define BOOST_PP_BOOL_796 1
# define BOOST_PP_BOOL_797 1
# define BOOST_PP_BOOL_798 1
# define BOOST_PP_BOOL_799 1
# define BOOST_PP_BOOL_800 1
# define BOOST_PP_BOOL_801 1
# define BOOST_PP_BOOL_802 1
# define BOOST_PP_BOOL_803 1
# define BOOST_PP_BOOL_804 1
# define BOOST_PP_BOOL_805 1
# define BOOST_PP_BOOL_806 1
# define BOOST_PP_BOOL_807 1
# define BOOST_PP_BOOL_808 1
# define BOOST_PP_BOOL_809 1
# define BOOST_PP_BOOL_810 1
# define BOOST_PP_BOOL_811 1
# define BOOST_PP_BOOL_812 1
# define BOOST_PP_BOOL_813 1
# define BOOST_PP_BOOL_814 1
# define BOOST_PP_BOOL_815 1
# define BOOST_PP_BOOL_816 1
# define BOOST_PP_BOOL_817 1
# define BOOST_PP_BOOL_818 1
# define BOOST_PP_BOOL_819 1
# define BOOST_PP_BOOL_820 1
# define BOOST_PP_BOOL_821 1
# define BOOST_PP_BOOL_822 1
# define BOOST_PP_BOOL_823 1
# define BOOST_PP_BOOL_824 1
# define BOOST_PP_BOOL_825 1
# define BOOST_PP_BOOL_826 1
# define BOOST_PP_BOOL_827 1
# define BOOST_PP_BOOL_828 1
# define BOOST_PP_BOOL_829 1
# define BOOST_PP_BOOL_830 1
# define BOOST_PP_BOOL_831 1
# define BOOST_PP_BOOL_832 1
# define BOOST_PP_BOOL_833 1
# define BOOST_PP_BOOL_834 1
# define BOOST_PP_BOOL_835 1
# define BOOST_PP_BOOL_836 1
# define BOOST_PP_BOOL_837 1
# define BOOST_PP_BOOL_838 1
# define BOOST_PP_BOOL_839 1
# define BOOST_PP_BOOL_840 1
# define BOOST_PP_BOOL_841 1
# define BOOST_PP_BOOL_842 1
# define BOOST_PP_BOOL_843 1
# define BOOST_PP_BOOL_844 1
# define BOOST_PP_BOOL_845 1
# define BOOST_PP_BOOL_846 1
# define BOOST_PP_BOOL_847 1
# define BOOST_PP_BOOL_848 1
# define BOOST_PP_BOOL_849 1
# define BOOST_PP_BOOL_850 1
# define BOOST_PP_BOOL_851 1
# define BOOST_PP_BOOL_852 1
# define BOOST_PP_BOOL_853 1
# define BOOST_PP_BOOL_854 1
# define BOOST_PP_BOOL_855 1
# define BOOST_PP_BOOL_856 1
# define BOOST_PP_BOOL_857 1
# define BOOST_PP_BOOL_858 1
# define BOOST_PP_BOOL_859 1
# define BOOST_PP_BOOL_860 1
# define BOOST_PP_BOOL_861 1
# define BOOST_PP_BOOL_862 1
# define BOOST_PP_BOOL_863 1
# define BOOST_PP_BOOL_864 1
# define BOOST_PP_BOOL_865 1
# define BOOST_PP_BOOL_866 1
# define BOOST_PP_BOOL_867 1
# define BOOST_PP_BOOL_868 1
# define BOOST_PP_BOOL_869 1
# define BOOST_PP_BOOL_870 1
# define BOOST_PP_BOOL_871 1
# define BOOST_PP_BOOL_872 1
# define BOOST_PP_BOOL_873 1
# define BOOST_PP_BOOL_874 1
# define BOOST_PP_BOOL_875 1
# define BOOST_PP_BOOL_876 1
# define BOOST_PP_BOOL_877 1
# define BOOST_PP_BOOL_878 1
# define BOOST_PP_BOOL_879 1
# define BOOST_PP_BOOL_880 1
# define BOOST_PP_BOOL_881 1
# define BOOST_PP_BOOL_882 1
# define BOOST_PP_BOOL_883 1
# define BOOST_PP_BOOL_884 1
# define BOOST_PP_BOOL_885 1
# define BOOST_PP_BOOL_886 1
# define BOOST_PP_BOOL_887 1
# define BOOST_PP_BOOL_888 1
# define BOOST_PP_BOOL_889 1
# define BOOST_PP_BOOL_890 1
# define BOOST_PP_BOOL_891 1
# define BOOST_PP_BOOL_892 1
# define BOOST_PP_BOOL_893 1
# define BOOST_PP_BOOL_894 1
# define BOOST_PP_BOOL_895 1
# define BOOST_PP_BOOL_896 1
# define BOOST_PP_BOOL_897 1
# define BOOST_PP_BOOL_898 1
# define BOOST_PP_BOOL_899 1
# define BOOST_PP_BOOL_900 1
# define BOOST_PP_BOOL_901 1
# define BOOST_PP_BOOL_902 1
# define BOOST_PP_BOOL_903 1
# define BOOST_PP_BOOL_904 1
# define BOOST_PP_BOOL_905 1
# define BOOST_PP_BOOL_906 1
# define BOOST_PP_BOOL_907 1
# define BOOST_PP_BOOL_908 1
# define BOOST_PP_BOOL_909 1
# define BOOST_PP_BOOL_910 1
# define BOOST_PP_BOOL_911 1
# define BOOST_PP_BOOL_912 1
# define BOOST_PP_BOOL_913 1
# define BOOST_PP_BOOL_914 1
# define BOOST_PP_BOOL_915 1
# define BOOST_PP_BOOL_916 1
# define BOOST_PP_BOOL_917 1
# define BOOST_PP_BOOL_918 1
# define BOOST_PP_BOOL_919 1
# define BOOST_PP_BOOL_920 1
# define BOOST_PP_BOOL_921 1
# define BOOST_PP_BOOL_922 1
# define BOOST_PP_BOOL_923 1
# define BOOST_PP_BOOL_924 1
# define BOOST_PP_BOOL_925 1
# define BOOST_PP_BOOL_926 1
# define BOOST_PP_BOOL_927 1
# define BOOST_PP_BOOL_928 1
# define BOOST_PP_BOOL_929 1
# define BOOST_PP_BOOL_930 1
# define BOOST_PP_BOOL_931 1
# define BOOST_PP_BOOL_932 1
# define BOOST_PP_BOOL_933 1
# define BOOST_PP_BOOL_934 1
# define BOOST_PP_BOOL_935 1
# define BOOST_PP_BOOL_936 1
# define BOOST_PP_BOOL_937 1
# define BOOST_PP_BOOL_938 1
# define BOOST_PP_BOOL_939 1
# define BOOST_PP_BOOL_940 1
# define BOOST_PP_BOOL_941 1
# define BOOST_PP_BOOL_942 1
# define BOOST_PP_BOOL_943 1
# define BOOST_PP_BOOL_944 1
# define BOOST_PP_BOOL_945 1
# define BOOST_PP_BOOL_946 1
# define BOOST_PP_BOOL_947 1
# define BOOST_PP_BOOL_948 1
# define BOOST_PP_BOOL_949 1
# define BOOST_PP_BOOL_950 1
# define BOOST_PP_BOOL_951 1
# define BOOST_PP_BOOL_952 1
# define BOOST_PP_BOOL_953 1
# define BOOST_PP_BOOL_954 1
# define BOOST_PP_BOOL_955 1
# define BOOST_PP_BOOL_956 1
# define BOOST_PP_BOOL_957 1
# define BOOST_PP_BOOL_958 1
# define BOOST_PP_BOOL_959 1
# define BOOST_PP_BOOL_960 1
# define BOOST_PP_BOOL_961 1
# define BOOST_PP_BOOL_962 1
# define BOOST_PP_BOOL_963 1
# define BOOST_PP_BOOL_964 1
# define BOOST_PP_BOOL_965 1
# define BOOST_PP_BOOL_966 1
# define BOOST_PP_BOOL_967 1
# define BOOST_PP_BOOL_968 1
# define BOOST_PP_BOOL_969 1
# define BOOST_PP_BOOL_970 1
# define BOOST_PP_BOOL_971 1
# define BOOST_PP_BOOL_972 1
# define BOOST_PP_BOOL_973 1
# define BOOST_PP_BOOL_974 1
# define BOOST_PP_BOOL_975 1
# define BOOST_PP_BOOL_976 1
# define BOOST_PP_BOOL_977 1
# define BOOST_PP_BOOL_978 1
# define BOOST_PP_BOOL_979 1
# define BOOST_PP_BOOL_980 1
# define BOOST_PP_BOOL_981 1
# define BOOST_PP_BOOL_982 1
# define BOOST_PP_BOOL_983 1
# define BOOST_PP_BOOL_984 1
# define BOOST_PP_BOOL_985 1
# define BOOST_PP_BOOL_986 1
# define BOOST_PP_BOOL_987 1
# define BOOST_PP_BOOL_988 1
# define BOOST_PP_BOOL_989 1
# define BOOST_PP_BOOL_990 1
# define BOOST_PP_BOOL_991 1
# define BOOST_PP_BOOL_992 1
# define BOOST_PP_BOOL_993 1
# define BOOST_PP_BOOL_994 1
# define BOOST_PP_BOOL_995 1
# define BOOST_PP_BOOL_996 1
# define BOOST_PP_BOOL_997 1
# define BOOST_PP_BOOL_998 1
# define BOOST_PP_BOOL_999 1
# define BOOST_PP_BOOL_1000 1
# define BOOST_PP_BOOL_1001 1
# define BOOST_PP_BOOL_1002 1
# define BOOST_PP_BOOL_1003 1
# define BOOST_PP_BOOL_1004 1
# define BOOST_PP_BOOL_1005 1
# define BOOST_PP_BOOL_1006 1
# define BOOST_PP_BOOL_1007 1
# define BOOST_PP_BOOL_1008 1
# define BOOST_PP_BOOL_1009 1
# define BOOST_PP_BOOL_1010 1
# define BOOST_PP_BOOL_1011 1
# define BOOST_PP_BOOL_1012 1
# define BOOST_PP_BOOL_1013 1
# define BOOST_PP_BOOL_1014 1
# define BOOST_PP_BOOL_1015 1
# define BOOST_PP_BOOL_1016 1
# define BOOST_PP_BOOL_1017 1
# define BOOST_PP_BOOL_1018 1
# define BOOST_PP_BOOL_1019 1
# define BOOST_PP_BOOL_1020 1
# define BOOST_PP_BOOL_1021 1
# define BOOST_PP_BOOL_1022 1
# define BOOST_PP_BOOL_1023 1
# define BOOST_PP_BOOL_1024 1
#
# endif
