# /* **************************************************************************
#  *                                                                          *
#  *     (C) Copyright Paul Mensonides 2002.
#  *     Distributed under the Boost Software License, Version 1.0. (See
#  *     accompanying file LICENSE_1_0.txt or copy at
#  *     http://www.boost.org/LICENSE_1_0.txt)
#  *                                                                          *
#  ************************************************************************** */
#
# /* Revised by Edward Diener (2020) */
#
# /* See http://www.boost.org for most recent version. */
#
# ifndef BOOST_PREPROCESSOR_SEQ_ENUM_1024_HPP
# define BOOST_PREPROCESSOR_SEQ_ENUM_1024_HPP
#
# define BOOST_PP_SEQ_ENUM_513(x) x, BOOST_PP_SEQ_ENUM_512
# define BOOST_PP_SEQ_ENUM_514(x) x, BOOST_PP_SEQ_ENUM_513
# define BOOST_PP_SEQ_ENUM_515(x) x, BOOST_PP_SEQ_ENUM_514
# define BOOST_PP_SEQ_ENUM_516(x) x, BOOST_PP_SEQ_ENUM_515
# define BOOST_PP_SEQ_ENUM_517(x) x, BOOST_PP_SEQ_ENUM_516
# define BOOST_PP_SEQ_ENUM_518(x) x, BOOST_PP_SEQ_ENUM_517
# define BOOST_PP_SEQ_ENUM_519(x) x, BOOST_PP_SEQ_ENUM_518
# define BOOST_PP_SEQ_ENUM_520(x) x, BOOST_PP_SEQ_ENUM_519
# define BOOST_PP_SEQ_ENUM_521(x) x, BOOST_PP_SEQ_ENUM_520
# define BOOST_PP_SEQ_ENUM_522(x) x, BOOST_PP_SEQ_ENUM_521
# define BOOST_PP_SEQ_ENUM_523(x) x, BOOST_PP_SEQ_ENUM_522
# define BOOST_PP_SEQ_ENUM_524(x) x, BOOST_PP_SEQ_ENUM_523
# define BOOST_PP_SEQ_ENUM_525(x) x, BOOST_PP_SEQ_ENUM_524
# define BOOST_PP_SEQ_ENUM_526(x) x, BOOST_PP_SEQ_ENUM_525
# define BOOST_PP_SEQ_ENUM_527(x) x, BOOST_PP_SEQ_ENUM_526
# define BOOST_PP_SEQ_ENUM_528(x) x, BOOST_PP_SEQ_ENUM_527
# define BOOST_PP_SEQ_ENUM_529(x) x, BOOST_PP_SEQ_ENUM_528
# define BOOST_PP_SEQ_ENUM_530(x) x, BOOST_PP_SEQ_ENUM_529
# define BOOST_PP_SEQ_ENUM_531(x) x, BOOST_PP_SEQ_ENUM_530
# define BOOST_PP_SEQ_ENUM_532(x) x, BOOST_PP_SEQ_ENUM_531
# define BOOST_PP_SEQ_ENUM_533(x) x, BOOST_PP_SEQ_ENUM_532
# define BOOST_PP_SEQ_ENUM_534(x) x, BOOST_PP_SEQ_ENUM_533
# define BOOST_PP_SEQ_ENUM_535(x) x, BOOST_PP_SEQ_ENUM_534
# define BOOST_PP_SEQ_ENUM_536(x) x, BOOST_PP_SEQ_ENUM_535
# define BOOST_PP_SEQ_ENUM_537(x) x, BOOST_PP_SEQ_ENUM_536
# define BOOST_PP_SEQ_ENUM_538(x) x, BOOST_PP_SEQ_ENUM_537
# define BOOST_PP_SEQ_ENUM_539(x) x, BOOST_PP_SEQ_ENUM_538
# define BOOST_PP_SEQ_ENUM_540(x) x, BOOST_PP_SEQ_ENUM_539
# define BOOST_PP_SEQ_ENUM_541(x) x, BOOST_PP_SEQ_ENUM_540
# define BOOST_PP_SEQ_ENUM_542(x) x, BOOST_PP_SEQ_ENUM_541
# define BOOST_PP_SEQ_ENUM_543(x) x, BOOST_PP_SEQ_ENUM_542
# define BOOST_PP_SEQ_ENUM_544(x) x, BOOST_PP_SEQ_ENUM_543
# define BOOST_PP_SEQ_ENUM_545(x) x, BOOST_PP_SEQ_ENUM_544
# define BOOST_PP_SEQ_ENUM_546(x) x, BOOST_PP_SEQ_ENUM_545
# define BOOST_PP_SEQ_ENUM_547(x) x, BOOST_PP_SEQ_ENUM_546
# define BOOST_PP_SEQ_ENUM_548(x) x, BOOST_PP_SEQ_ENUM_547
# define BOOST_PP_SEQ_ENUM_549(x) x, BOOST_PP_SEQ_ENUM_548
# define BOOST_PP_SEQ_ENUM_550(x) x, BOOST_PP_SEQ_ENUM_549
# define BOOST_PP_SEQ_ENUM_551(x) x, BOOST_PP_SEQ_ENUM_550
# define BOOST_PP_SEQ_ENUM_552(x) x, BOOST_PP_SEQ_ENUM_551
# define BOOST_PP_SEQ_ENUM_553(x) x, BOOST_PP_SEQ_ENUM_552
# define BOOST_PP_SEQ_ENUM_554(x) x, BOOST_PP_SEQ_ENUM_553
# define BOOST_PP_SEQ_ENUM_555(x) x, BOOST_PP_SEQ_ENUM_554
# define BOOST_PP_SEQ_ENUM_556(x) x, BOOST_PP_SEQ_ENUM_555
# define BOOST_PP_SEQ_ENUM_557(x) x, BOOST_PP_SEQ_ENUM_556
# define BOOST_PP_SEQ_ENUM_558(x) x, BOOST_PP_SEQ_ENUM_557
# define BOOST_PP_SEQ_ENUM_559(x) x, BOOST_PP_SEQ_ENUM_558
# define BOOST_PP_SEQ_ENUM_560(x) x, BOOST_PP_SEQ_ENUM_559
# define BOOST_PP_SEQ_ENUM_561(x) x, BOOST_PP_SEQ_ENUM_560
# define BOOST_PP_SEQ_ENUM_562(x) x, BOOST_PP_SEQ_ENUM_561
# define BOOST_PP_SEQ_ENUM_563(x) x, BOOST_PP_SEQ_ENUM_562
# define BOOST_PP_SEQ_ENUM_564(x) x, BOOST_PP_SEQ_ENUM_563
# define BOOST_PP_SEQ_ENUM_565(x) x, BOOST_PP_SEQ_ENUM_564
# define BOOST_PP_SEQ_ENUM_566(x) x, BOOST_PP_SEQ_ENUM_565
# define BOOST_PP_SEQ_ENUM_567(x) x, BOOST_PP_SEQ_ENUM_566
# define BOOST_PP_SEQ_ENUM_568(x) x, BOOST_PP_SEQ_ENUM_567
# define BOOST_PP_SEQ_ENUM_569(x) x, BOOST_PP_SEQ_ENUM_568
# define BOOST_PP_SEQ_ENUM_570(x) x, BOOST_PP_SEQ_ENUM_569
# define BOOST_PP_SEQ_ENUM_571(x) x, BOOST_PP_SEQ_ENUM_570
# define BOOST_PP_SEQ_ENUM_572(x) x, BOOST_PP_SEQ_ENUM_571
# define BOOST_PP_SEQ_ENUM_573(x) x, BOOST_PP_SEQ_ENUM_572
# define BOOST_PP_SEQ_ENUM_574(x) x, BOOST_PP_SEQ_ENUM_573
# define BOOST_PP_SEQ_ENUM_575(x) x, BOOST_PP_SEQ_ENUM_574
# define BOOST_PP_SEQ_ENUM_576(x) x, BOOST_PP_SEQ_ENUM_575
# define BOOST_PP_SEQ_ENUM_577(x) x, BOOST_PP_SEQ_ENUM_576
# define BOOST_PP_SEQ_ENUM_578(x) x, BOOST_PP_SEQ_ENUM_577
# define BOOST_PP_SEQ_ENUM_579(x) x, BOOST_PP_SEQ_ENUM_578
# define BOOST_PP_SEQ_ENUM_580(x) x, BOOST_PP_SEQ_ENUM_579
# define BOOST_PP_SEQ_ENUM_581(x) x, BOOST_PP_SEQ_ENUM_580
# define BOOST_PP_SEQ_ENUM_582(x) x, BOOST_PP_SEQ_ENUM_581
# define BOOST_PP_SEQ_ENUM_583(x) x, BOOST_PP_SEQ_ENUM_582
# define BOOST_PP_SEQ_ENUM_584(x) x, BOOST_PP_SEQ_ENUM_583
# define BOOST_PP_SEQ_ENUM_585(x) x, BOOST_PP_SEQ_ENUM_584
# define BOOST_PP_SEQ_ENUM_586(x) x, BOOST_PP_SEQ_ENUM_585
# define BOOST_PP_SEQ_ENUM_587(x) x, BOOST_PP_SEQ_ENUM_586
# define BOOST_PP_SEQ_ENUM_588(x) x, BOOST_PP_SEQ_ENUM_587
# define BOOST_PP_SEQ_ENUM_589(x) x, BOOST_PP_SEQ_ENUM_588
# define BOOST_PP_SEQ_ENUM_590(x) x, BOOST_PP_SEQ_ENUM_589
# define BOOST_PP_SEQ_ENUM_591(x) x, BOOST_PP_SEQ_ENUM_590
# define BOOST_PP_SEQ_ENUM_592(x) x, BOOST_PP_SEQ_ENUM_591
# define BOOST_PP_SEQ_ENUM_593(x) x, BOOST_PP_SEQ_ENUM_592
# define BOOST_PP_SEQ_ENUM_594(x) x, BOOST_PP_SEQ_ENUM_593
# define BOOST_PP_SEQ_ENUM_595(x) x, BOOST_PP_SEQ_ENUM_594
# define BOOST_PP_SEQ_ENUM_596(x) x, BOOST_PP_SEQ_ENUM_595
# define BOOST_PP_SEQ_ENUM_597(x) x, BOOST_PP_SEQ_ENUM_596
# define BOOST_PP_SEQ_ENUM_598(x) x, BOOST_PP_SEQ_ENUM_597
# define BOOST_PP_SEQ_ENUM_599(x) x, BOOST_PP_SEQ_ENUM_598
# define BOOST_PP_SEQ_ENUM_600(x) x, BOOST_PP_SEQ_ENUM_599
# define BOOST_PP_SEQ_ENUM_601(x) x, BOOST_PP_SEQ_ENUM_600
# define BOOST_PP_SEQ_ENUM_602(x) x, BOOST_PP_SEQ_ENUM_601
# define BOOST_PP_SEQ_ENUM_603(x) x, BOOST_PP_SEQ_ENUM_602
# define BOOST_PP_SEQ_ENUM_604(x) x, BOOST_PP_SEQ_ENUM_603
# define BOOST_PP_SEQ_ENUM_605(x) x, BOOST_PP_SEQ_ENUM_604
# define BOOST_PP_SEQ_ENUM_606(x) x, BOOST_PP_SEQ_ENUM_605
# define BOOST_PP_SEQ_ENUM_607(x) x, BOOST_PP_SEQ_ENUM_606
# define BOOST_PP_SEQ_ENUM_608(x) x, BOOST_PP_SEQ_ENUM_607
# define BOOST_PP_SEQ_ENUM_609(x) x, BOOST_PP_SEQ_ENUM_608
# define BOOST_PP_SEQ_ENUM_610(x) x, BOOST_PP_SEQ_ENUM_609
# define BOOST_PP_SEQ_ENUM_611(x) x, BOOST_PP_SEQ_ENUM_610
# define BOOST_PP_SEQ_ENUM_612(x) x, BOOST_PP_SEQ_ENUM_611
# define BOOST_PP_SEQ_ENUM_613(x) x, BOOST_PP_SEQ_ENUM_612
# define BOOST_PP_SEQ_ENUM_614(x) x, BOOST_PP_SEQ_ENUM_613
# define BOOST_PP_SEQ_ENUM_615(x) x, BOOST_PP_SEQ_ENUM_614
# define BOOST_PP_SEQ_ENUM_616(x) x, BOOST_PP_SEQ_ENUM_615
# define BOOST_PP_SEQ_ENUM_617(x) x, BOOST_PP_SEQ_ENUM_616
# define BOOST_PP_SEQ_ENUM_618(x) x, BOOST_PP_SEQ_ENUM_617
# define BOOST_PP_SEQ_ENUM_619(x) x, BOOST_PP_SEQ_ENUM_618
# define BOOST_PP_SEQ_ENUM_620(x) x, BOOST_PP_SEQ_ENUM_619
# define BOOST_PP_SEQ_ENUM_621(x) x, BOOST_PP_SEQ_ENUM_620
# define BOOST_PP_SEQ_ENUM_622(x) x, BOOST_PP_SEQ_ENUM_621
# define BOOST_PP_SEQ_ENUM_623(x) x, BOOST_PP_SEQ_ENUM_622
# define BOOST_PP_SEQ_ENUM_624(x) x, BOOST_PP_SEQ_ENUM_623
# define BOOST_PP_SEQ_ENUM_625(x) x, BOOST_PP_SEQ_ENUM_624
# define BOOST_PP_SEQ_ENUM_626(x) x, BOOST_PP_SEQ_ENUM_625
# define BOOST_PP_SEQ_ENUM_627(x) x, BOOST_PP_SEQ_ENUM_626
# define BOOST_PP_SEQ_ENUM_628(x) x, BOOST_PP_SEQ_ENUM_627
# define BOOST_PP_SEQ_ENUM_629(x) x, BOOST_PP_SEQ_ENUM_628
# define BOOST_PP_SEQ_ENUM_630(x) x, BOOST_PP_SEQ_ENUM_629
# define BOOST_PP_SEQ_ENUM_631(x) x, BOOST_PP_SEQ_ENUM_630
# define BOOST_PP_SEQ_ENUM_632(x) x, BOOST_PP_SEQ_ENUM_631
# define BOOST_PP_SEQ_ENUM_633(x) x, BOOST_PP_SEQ_ENUM_632
# define BOOST_PP_SEQ_ENUM_634(x) x, BOOST_PP_SEQ_ENUM_633
# define BOOST_PP_SEQ_ENUM_635(x) x, BOOST_PP_SEQ_ENUM_634
# define BOOST_PP_SEQ_ENUM_636(x) x, BOOST_PP_SEQ_ENUM_635
# define BOOST_PP_SEQ_ENUM_637(x) x, BOOST_PP_SEQ_ENUM_636
# define BOOST_PP_SEQ_ENUM_638(x) x, BOOST_PP_SEQ_ENUM_637
# define BOOST_PP_SEQ_ENUM_639(x) x, BOOST_PP_SEQ_ENUM_638
# define BOOST_PP_SEQ_ENUM_640(x) x, BOOST_PP_SEQ_ENUM_639
# define BOOST_PP_SEQ_ENUM_641(x) x, BOOST_PP_SEQ_ENUM_640
# define BOOST_PP_SEQ_ENUM_642(x) x, BOOST_PP_SEQ_ENUM_641
# define BOOST_PP_SEQ_ENUM_643(x) x, BOOST_PP_SEQ_ENUM_642
# define BOOST_PP_SEQ_ENUM_644(x) x, BOOST_PP_SEQ_ENUM_643
# define BOOST_PP_SEQ_ENUM_645(x) x, BOOST_PP_SEQ_ENUM_644
# define BOOST_PP_SEQ_ENUM_646(x) x, BOOST_PP_SEQ_ENUM_645
# define BOOST_PP_SEQ_ENUM_647(x) x, BOOST_PP_SEQ_ENUM_646
# define BOOST_PP_SEQ_ENUM_648(x) x, BOOST_PP_SEQ_ENUM_647
# define BOOST_PP_SEQ_ENUM_649(x) x, BOOST_PP_SEQ_ENUM_648
# define BOOST_PP_SEQ_ENUM_650(x) x, BOOST_PP_SEQ_ENUM_649
# define BOOST_PP_SEQ_ENUM_651(x) x, BOOST_PP_SEQ_ENUM_650
# define BOOST_PP_SEQ_ENUM_652(x) x, BOOST_PP_SEQ_ENUM_651
# define BOOST_PP_SEQ_ENUM_653(x) x, BOOST_PP_SEQ_ENUM_652
# define BOOST_PP_SEQ_ENUM_654(x) x, BOOST_PP_SEQ_ENUM_653
# define BOOST_PP_SEQ_ENUM_655(x) x, BOOST_PP_SEQ_ENUM_654
# define BOOST_PP_SEQ_ENUM_656(x) x, BOOST_PP_SEQ_ENUM_655
# define BOOST_PP_SEQ_ENUM_657(x) x, BOOST_PP_SEQ_ENUM_656
# define BOOST_PP_SEQ_ENUM_658(x) x, BOOST_PP_SEQ_ENUM_657
# define BOOST_PP_SEQ_ENUM_659(x) x, BOOST_PP_SEQ_ENUM_658
# define BOOST_PP_SEQ_ENUM_660(x) x, BOOST_PP_SEQ_ENUM_659
# define BOOST_PP_SEQ_ENUM_661(x) x, BOOST_PP_SEQ_ENUM_660
# define BOOST_PP_SEQ_ENUM_662(x) x, BOOST_PP_SEQ_ENUM_661
# define BOOST_PP_SEQ_ENUM_663(x) x, BOOST_PP_SEQ_ENUM_662
# define BOOST_PP_SEQ_ENUM_664(x) x, BOOST_PP_SEQ_ENUM_663
# define BOOST_PP_SEQ_ENUM_665(x) x, BOOST_PP_SEQ_ENUM_664
# define BOOST_PP_SEQ_ENUM_666(x) x, BOOST_PP_SEQ_ENUM_665
# define BOOST_PP_SEQ_ENUM_667(x) x, BOOST_PP_SEQ_ENUM_666
# define BOOST_PP_SEQ_ENUM_668(x) x, BOOST_PP_SEQ_ENUM_667
# define BOOST_PP_SEQ_ENUM_669(x) x, BOOST_PP_SEQ_ENUM_668
# define BOOST_PP_SEQ_ENUM_670(x) x, BOOST_PP_SEQ_ENUM_669
# define BOOST_PP_SEQ_ENUM_671(x) x, BOOST_PP_SEQ_ENUM_670
# define BOOST_PP_SEQ_ENUM_672(x) x, BOOST_PP_SEQ_ENUM_671
# define BOOST_PP_SEQ_ENUM_673(x) x, BOOST_PP_SEQ_ENUM_672
# define BOOST_PP_SEQ_ENUM_674(x) x, BOOST_PP_SEQ_ENUM_673
# define BOOST_PP_SEQ_ENUM_675(x) x, BOOST_PP_SEQ_ENUM_674
# define BOOST_PP_SEQ_ENUM_676(x) x, BOOST_PP_SEQ_ENUM_675
# define BOOST_PP_SEQ_ENUM_677(x) x, BOOST_PP_SEQ_ENUM_676
# define BOOST_PP_SEQ_ENUM_678(x) x, BOOST_PP_SEQ_ENUM_677
# define BOOST_PP_SEQ_ENUM_679(x) x, BOOST_PP_SEQ_ENUM_678
# define BOOST_PP_SEQ_ENUM_680(x) x, BOOST_PP_SEQ_ENUM_679
# define BOOST_PP_SEQ_ENUM_681(x) x, BOOST_PP_SEQ_ENUM_680
# define BOOST_PP_SEQ_ENUM_682(x) x, BOOST_PP_SEQ_ENUM_681
# define BOOST_PP_SEQ_ENUM_683(x) x, BOOST_PP_SEQ_ENUM_682
# define BOOST_PP_SEQ_ENUM_684(x) x, BOOST_PP_SEQ_ENUM_683
# define BOOST_PP_SEQ_ENUM_685(x) x, BOOST_PP_SEQ_ENUM_684
# define BOOST_PP_SEQ_ENUM_686(x) x, BOOST_PP_SEQ_ENUM_685
# define BOOST_PP_SEQ_ENUM_687(x) x, BOOST_PP_SEQ_ENUM_686
# define BOOST_PP_SEQ_ENUM_688(x) x, BOOST_PP_SEQ_ENUM_687
# define BOOST_PP_SEQ_ENUM_689(x) x, BOOST_PP_SEQ_ENUM_688
# define BOOST_PP_SEQ_ENUM_690(x) x, BOOST_PP_SEQ_ENUM_689
# define BOOST_PP_SEQ_ENUM_691(x) x, BOOST_PP_SEQ_ENUM_690
# define BOOST_PP_SEQ_ENUM_692(x) x, BOOST_PP_SEQ_ENUM_691
# define BOOST_PP_SEQ_ENUM_693(x) x, BOOST_PP_SEQ_ENUM_692
# define BOOST_PP_SEQ_ENUM_694(x) x, BOOST_PP_SEQ_ENUM_693
# define BOOST_PP_SEQ_ENUM_695(x) x, BOOST_PP_SEQ_ENUM_694
# define BOOST_PP_SEQ_ENUM_696(x) x, BOOST_PP_SEQ_ENUM_695
# define BOOST_PP_SEQ_ENUM_697(x) x, BOOST_PP_SEQ_ENUM_696
# define BOOST_PP_SEQ_ENUM_698(x) x, BOOST_PP_SEQ_ENUM_697
# define BOOST_PP_SEQ_ENUM_699(x) x, BOOST_PP_SEQ_ENUM_698
# define BOOST_PP_SEQ_ENUM_700(x) x, BOOST_PP_SEQ_ENUM_699
# define BOOST_PP_SEQ_ENUM_701(x) x, BOOST_PP_SEQ_ENUM_700
# define BOOST_PP_SEQ_ENUM_702(x) x, BOOST_PP_SEQ_ENUM_701
# define BOOST_PP_SEQ_ENUM_703(x) x, BOOST_PP_SEQ_ENUM_702
# define BOOST_PP_SEQ_ENUM_704(x) x, BOOST_PP_SEQ_ENUM_703
# define BOOST_PP_SEQ_ENUM_705(x) x, BOOST_PP_SEQ_ENUM_704
# define BOOST_PP_SEQ_ENUM_706(x) x, BOOST_PP_SEQ_ENUM_705
# define BOOST_PP_SEQ_ENUM_707(x) x, BOOST_PP_SEQ_ENUM_706
# define BOOST_PP_SEQ_ENUM_708(x) x, BOOST_PP_SEQ_ENUM_707
# define BOOST_PP_SEQ_ENUM_709(x) x, BOOST_PP_SEQ_ENUM_708
# define BOOST_PP_SEQ_ENUM_710(x) x, BOOST_PP_SEQ_ENUM_709
# define BOOST_PP_SEQ_ENUM_711(x) x, BOOST_PP_SEQ_ENUM_710
# define BOOST_PP_SEQ_ENUM_712(x) x, BOOST_PP_SEQ_ENUM_711
# define BOOST_PP_SEQ_ENUM_713(x) x, BOOST_PP_SEQ_ENUM_712
# define BOOST_PP_SEQ_ENUM_714(x) x, BOOST_PP_SEQ_ENUM_713
# define BOOST_PP_SEQ_ENUM_715(x) x, BOOST_PP_SEQ_ENUM_714
# define BOOST_PP_SEQ_ENUM_716(x) x, BOOST_PP_SEQ_ENUM_715
# define BOOST_PP_SEQ_ENUM_717(x) x, BOOST_PP_SEQ_ENUM_716
# define BOOST_PP_SEQ_ENUM_718(x) x, BOOST_PP_SEQ_ENUM_717
# define BOOST_PP_SEQ_ENUM_719(x) x, BOOST_PP_SEQ_ENUM_718
# define BOOST_PP_SEQ_ENUM_720(x) x, BOOST_PP_SEQ_ENUM_719
# define BOOST_PP_SEQ_ENUM_721(x) x, BOOST_PP_SEQ_ENUM_720
# define BOOST_PP_SEQ_ENUM_722(x) x, BOOST_PP_SEQ_ENUM_721
# define BOOST_PP_SEQ_ENUM_723(x) x, BOOST_PP_SEQ_ENUM_722
# define BOOST_PP_SEQ_ENUM_724(x) x, BOOST_PP_SEQ_ENUM_723
# define BOOST_PP_SEQ_ENUM_725(x) x, BOOST_PP_SEQ_ENUM_724
# define BOOST_PP_SEQ_ENUM_726(x) x, BOOST_PP_SEQ_ENUM_725
# define BOOST_PP_SEQ_ENUM_727(x) x, BOOST_PP_SEQ_ENUM_726
# define BOOST_PP_SEQ_ENUM_728(x) x, BOOST_PP_SEQ_ENUM_727
# define BOOST_PP_SEQ_ENUM_729(x) x, BOOST_PP_SEQ_ENUM_728
# define BOOST_PP_SEQ_ENUM_730(x) x, BOOST_PP_SEQ_ENUM_729
# define BOOST_PP_SEQ_ENUM_731(x) x, BOOST_PP_SEQ_ENUM_730
# define BOOST_PP_SEQ_ENUM_732(x) x, BOOST_PP_SEQ_ENUM_731
# define BOOST_PP_SEQ_ENUM_733(x) x, BOOST_PP_SEQ_ENUM_732
# define BOOST_PP_SEQ_ENUM_734(x) x, BOOST_PP_SEQ_ENUM_733
# define BOOST_PP_SEQ_ENUM_735(x) x, BOOST_PP_SEQ_ENUM_734
# define BOOST_PP_SEQ_ENUM_736(x) x, BOOST_PP_SEQ_ENUM_735
# define BOOST_PP_SEQ_ENUM_737(x) x, BOOST_PP_SEQ_ENUM_736
# define BOOST_PP_SEQ_ENUM_738(x) x, BOOST_PP_SEQ_ENUM_737
# define BOOST_PP_SEQ_ENUM_739(x) x, BOOST_PP_SEQ_ENUM_738
# define BOOST_PP_SEQ_ENUM_740(x) x, BOOST_PP_SEQ_ENUM_739
# define BOOST_PP_SEQ_ENUM_741(x) x, BOOST_PP_SEQ_ENUM_740
# define BOOST_PP_SEQ_ENUM_742(x) x, BOOST_PP_SEQ_ENUM_741
# define BOOST_PP_SEQ_ENUM_743(x) x, BOOST_PP_SEQ_ENUM_742
# define BOOST_PP_SEQ_ENUM_744(x) x, BOOST_PP_SEQ_ENUM_743
# define BOOST_PP_SEQ_ENUM_745(x) x, BOOST_PP_SEQ_ENUM_744
# define BOOST_PP_SEQ_ENUM_746(x) x, BOOST_PP_SEQ_ENUM_745
# define BOOST_PP_SEQ_ENUM_747(x) x, BOOST_PP_SEQ_ENUM_746
# define BOOST_PP_SEQ_ENUM_748(x) x, BOOST_PP_SEQ_ENUM_747
# define BOOST_PP_SEQ_ENUM_749(x) x, BOOST_PP_SEQ_ENUM_748
# define BOOST_PP_SEQ_ENUM_750(x) x, BOOST_PP_SEQ_ENUM_749
# define BOOST_PP_SEQ_ENUM_751(x) x, BOOST_PP_SEQ_ENUM_750
# define BOOST_PP_SEQ_ENUM_752(x) x, BOOST_PP_SEQ_ENUM_751
# define BOOST_PP_SEQ_ENUM_753(x) x, BOOST_PP_SEQ_ENUM_752
# define BOOST_PP_SEQ_ENUM_754(x) x, BOOST_PP_SEQ_ENUM_753
# define BOOST_PP_SEQ_ENUM_755(x) x, BOOST_PP_SEQ_ENUM_754
# define BOOST_PP_SEQ_ENUM_756(x) x, BOOST_PP_SEQ_ENUM_755
# define BOOST_PP_SEQ_ENUM_757(x) x, BOOST_PP_SEQ_ENUM_756
# define BOOST_PP_SEQ_ENUM_758(x) x, BOOST_PP_SEQ_ENUM_757
# define BOOST_PP_SEQ_ENUM_759(x) x, BOOST_PP_SEQ_ENUM_758
# define BOOST_PP_SEQ_ENUM_760(x) x, BOOST_PP_SEQ_ENUM_759
# define BOOST_PP_SEQ_ENUM_761(x) x, BOOST_PP_SEQ_ENUM_760
# define BOOST_PP_SEQ_ENUM_762(x) x, BOOST_PP_SEQ_ENUM_761
# define BOOST_PP_SEQ_ENUM_763(x) x, BOOST_PP_SEQ_ENUM_762
# define BOOST_PP_SEQ_ENUM_764(x) x, BOOST_PP_SEQ_ENUM_763
# define BOOST_PP_SEQ_ENUM_765(x) x, BOOST_PP_SEQ_ENUM_764
# define BOOST_PP_SEQ_ENUM_766(x) x, BOOST_PP_SEQ_ENUM_765
# define BOOST_PP_SEQ_ENUM_767(x) x, BOOST_PP_SEQ_ENUM_766
# define BOOST_PP_SEQ_ENUM_768(x) x, BOOST_PP_SEQ_ENUM_767
# define BOOST_PP_SEQ_ENUM_769(x) x, BOOST_PP_SEQ_ENUM_768
# define BOOST_PP_SEQ_ENUM_770(x) x, BOOST_PP_SEQ_ENUM_769
# define BOOST_PP_SEQ_ENUM_771(x) x, BOOST_PP_SEQ_ENUM_770
# define BOOST_PP_SEQ_ENUM_772(x) x, BOOST_PP_SEQ_ENUM_771
# define BOOST_PP_SEQ_ENUM_773(x) x, BOOST_PP_SEQ_ENUM_772
# define BOOST_PP_SEQ_ENUM_774(x) x, BOOST_PP_SEQ_ENUM_773
# define BOOST_PP_SEQ_ENUM_775(x) x, BOOST_PP_SEQ_ENUM_774
# define BOOST_PP_SEQ_ENUM_776(x) x, BOOST_PP_SEQ_ENUM_775
# define BOOST_PP_SEQ_ENUM_777(x) x, BOOST_PP_SEQ_ENUM_776
# define BOOST_PP_SEQ_ENUM_778(x) x, BOOST_PP_SEQ_ENUM_777
# define BOOST_PP_SEQ_ENUM_779(x) x, BOOST_PP_SEQ_ENUM_778
# define BOOST_PP_SEQ_ENUM_780(x) x, BOOST_PP_SEQ_ENUM_779
# define BOOST_PP_SEQ_ENUM_781(x) x, BOOST_PP_SEQ_ENUM_780
# define BOOST_PP_SEQ_ENUM_782(x) x, BOOST_PP_SEQ_ENUM_781
# define BOOST_PP_SEQ_ENUM_783(x) x, BOOST_PP_SEQ_ENUM_782
# define BOOST_PP_SEQ_ENUM_784(x) x, BOOST_PP_SEQ_ENUM_783
# define BOOST_PP_SEQ_ENUM_785(x) x, BOOST_PP_SEQ_ENUM_784
# define BOOST_PP_SEQ_ENUM_786(x) x, BOOST_PP_SEQ_ENUM_785
# define BOOST_PP_SEQ_ENUM_787(x) x, BOOST_PP_SEQ_ENUM_786
# define BOOST_PP_SEQ_ENUM_788(x) x, BOOST_PP_SEQ_ENUM_787
# define BOOST_PP_SEQ_ENUM_789(x) x, BOOST_PP_SEQ_ENUM_788
# define BOOST_PP_SEQ_ENUM_790(x) x, BOOST_PP_SEQ_ENUM_789
# define BOOST_PP_SEQ_ENUM_791(x) x, BOOST_PP_SEQ_ENUM_790
# define BOOST_PP_SEQ_ENUM_792(x) x, BOOST_PP_SEQ_ENUM_791
# define BOOST_PP_SEQ_ENUM_793(x) x, BOOST_PP_SEQ_ENUM_792
# define BOOST_PP_SEQ_ENUM_794(x) x, BOOST_PP_SEQ_ENUM_793
# define BOOST_PP_SEQ_ENUM_795(x) x, BOOST_PP_SEQ_ENUM_794
# define BOOST_PP_SEQ_ENUM_796(x) x, BOOST_PP_SEQ_ENUM_795
# define BOOST_PP_SEQ_ENUM_797(x) x, BOOST_PP_SEQ_ENUM_796
# define BOOST_PP_SEQ_ENUM_798(x) x, BOOST_PP_SEQ_ENUM_797
# define BOOST_PP_SEQ_ENUM_799(x) x, BOOST_PP_SEQ_ENUM_798
# define BOOST_PP_SEQ_ENUM_800(x) x, BOOST_PP_SEQ_ENUM_799
# define BOOST_PP_SEQ_ENUM_801(x) x, BOOST_PP_SEQ_ENUM_800
# define BOOST_PP_SEQ_ENUM_802(x) x, BOOST_PP_SEQ_ENUM_801
# define BOOST_PP_SEQ_ENUM_803(x) x, BOOST_PP_SEQ_ENUM_802
# define BOOST_PP_SEQ_ENUM_804(x) x, BOOST_PP_SEQ_ENUM_803
# define BOOST_PP_SEQ_ENUM_805(x) x, BOOST_PP_SEQ_ENUM_804
# define BOOST_PP_SEQ_ENUM_806(x) x, BOOST_PP_SEQ_ENUM_805
# define BOOST_PP_SEQ_ENUM_807(x) x, BOOST_PP_SEQ_ENUM_806
# define BOOST_PP_SEQ_ENUM_808(x) x, BOOST_PP_SEQ_ENUM_807
# define BOOST_PP_SEQ_ENUM_809(x) x, BOOST_PP_SEQ_ENUM_808
# define BOOST_PP_SEQ_ENUM_810(x) x, BOOST_PP_SEQ_ENUM_809
# define BOOST_PP_SEQ_ENUM_811(x) x, BOOST_PP_SEQ_ENUM_810
# define BOOST_PP_SEQ_ENUM_812(x) x, BOOST_PP_SEQ_ENUM_811
# define BOOST_PP_SEQ_ENUM_813(x) x, BOOST_PP_SEQ_ENUM_812
# define BOOST_PP_SEQ_ENUM_814(x) x, BOOST_PP_SEQ_ENUM_813
# define BOOST_PP_SEQ_ENUM_815(x) x, BOOST_PP_SEQ_ENUM_814
# define BOOST_PP_SEQ_ENUM_816(x) x, BOOST_PP_SEQ_ENUM_815
# define BOOST_PP_SEQ_ENUM_817(x) x, BOOST_PP_SEQ_ENUM_816
# define BOOST_PP_SEQ_ENUM_818(x) x, BOOST_PP_SEQ_ENUM_817
# define BOOST_PP_SEQ_ENUM_819(x) x, BOOST_PP_SEQ_ENUM_818
# define BOOST_PP_SEQ_ENUM_820(x) x, BOOST_PP_SEQ_ENUM_819
# define BOOST_PP_SEQ_ENUM_821(x) x, BOOST_PP_SEQ_ENUM_820
# define BOOST_PP_SEQ_ENUM_822(x) x, BOOST_PP_SEQ_ENUM_821
# define BOOST_PP_SEQ_ENUM_823(x) x, BOOST_PP_SEQ_ENUM_822
# define BOOST_PP_SEQ_ENUM_824(x) x, BOOST_PP_SEQ_ENUM_823
# define BOOST_PP_SEQ_ENUM_825(x) x, BOOST_PP_SEQ_ENUM_824
# define BOOST_PP_SEQ_ENUM_826(x) x, BOOST_PP_SEQ_ENUM_825
# define BOOST_PP_SEQ_ENUM_827(x) x, BOOST_PP_SEQ_ENUM_826
# define BOOST_PP_SEQ_ENUM_828(x) x, BOOST_PP_SEQ_ENUM_827
# define BOOST_PP_SEQ_ENUM_829(x) x, BOOST_PP_SEQ_ENUM_828
# define BOOST_PP_SEQ_ENUM_830(x) x, BOOST_PP_SEQ_ENUM_829
# define BOOST_PP_SEQ_ENUM_831(x) x, BOOST_PP_SEQ_ENUM_830
# define BOOST_PP_SEQ_ENUM_832(x) x, BOOST_PP_SEQ_ENUM_831
# define BOOST_PP_SEQ_ENUM_833(x) x, BOOST_PP_SEQ_ENUM_832
# define BOOST_PP_SEQ_ENUM_834(x) x, BOOST_PP_SEQ_ENUM_833
# define BOOST_PP_SEQ_ENUM_835(x) x, BOOST_PP_SEQ_ENUM_834
# define BOOST_PP_SEQ_ENUM_836(x) x, BOOST_PP_SEQ_ENUM_835
# define BOOST_PP_SEQ_ENUM_837(x) x, BOOST_PP_SEQ_ENUM_836
# define BOOST_PP_SEQ_ENUM_838(x) x, BOOST_PP_SEQ_ENUM_837
# define BOOST_PP_SEQ_ENUM_839(x) x, BOOST_PP_SEQ_ENUM_838
# define BOOST_PP_SEQ_ENUM_840(x) x, BOOST_PP_SEQ_ENUM_839
# define BOOST_PP_SEQ_ENUM_841(x) x, BOOST_PP_SEQ_ENUM_840
# define BOOST_PP_SEQ_ENUM_842(x) x, BOOST_PP_SEQ_ENUM_841
# define BOOST_PP_SEQ_ENUM_843(x) x, BOOST_PP_SEQ_ENUM_842
# define BOOST_PP_SEQ_ENUM_844(x) x, BOOST_PP_SEQ_ENUM_843
# define BOOST_PP_SEQ_ENUM_845(x) x, BOOST_PP_SEQ_ENUM_844
# define BOOST_PP_SEQ_ENUM_846(x) x, BOOST_PP_SEQ_ENUM_845
# define BOOST_PP_SEQ_ENUM_847(x) x, BOOST_PP_SEQ_ENUM_846
# define BOOST_PP_SEQ_ENUM_848(x) x, BOOST_PP_SEQ_ENUM_847
# define BOOST_PP_SEQ_ENUM_849(x) x, BOOST_PP_SEQ_ENUM_848
# define BOOST_PP_SEQ_ENUM_850(x) x, BOOST_PP_SEQ_ENUM_849
# define BOOST_PP_SEQ_ENUM_851(x) x, BOOST_PP_SEQ_ENUM_850
# define BOOST_PP_SEQ_ENUM_852(x) x, BOOST_PP_SEQ_ENUM_851
# define BOOST_PP_SEQ_ENUM_853(x) x, BOOST_PP_SEQ_ENUM_852
# define BOOST_PP_SEQ_ENUM_854(x) x, BOOST_PP_SEQ_ENUM_853
# define BOOST_PP_SEQ_ENUM_855(x) x, BOOST_PP_SEQ_ENUM_854
# define BOOST_PP_SEQ_ENUM_856(x) x, BOOST_PP_SEQ_ENUM_855
# define BOOST_PP_SEQ_ENUM_857(x) x, BOOST_PP_SEQ_ENUM_856
# define BOOST_PP_SEQ_ENUM_858(x) x, BOOST_PP_SEQ_ENUM_857
# define BOOST_PP_SEQ_ENUM_859(x) x, BOOST_PP_SEQ_ENUM_858
# define BOOST_PP_SEQ_ENUM_860(x) x, BOOST_PP_SEQ_ENUM_859
# define BOOST_PP_SEQ_ENUM_861(x) x, BOOST_PP_SEQ_ENUM_860
# define BOOST_PP_SEQ_ENUM_862(x) x, BOOST_PP_SEQ_ENUM_861
# define BOOST_PP_SEQ_ENUM_863(x) x, BOOST_PP_SEQ_ENUM_862
# define BOOST_PP_SEQ_ENUM_864(x) x, BOOST_PP_SEQ_ENUM_863
# define BOOST_PP_SEQ_ENUM_865(x) x, BOOST_PP_SEQ_ENUM_864
# define BOOST_PP_SEQ_ENUM_866(x) x, BOOST_PP_SEQ_ENUM_865
# define BOOST_PP_SEQ_ENUM_867(x) x, BOOST_PP_SEQ_ENUM_866
# define BOOST_PP_SEQ_ENUM_868(x) x, BOOST_PP_SEQ_ENUM_867
# define BOOST_PP_SEQ_ENUM_869(x) x, BOOST_PP_SEQ_ENUM_868
# define BOOST_PP_SEQ_ENUM_870(x) x, BOOST_PP_SEQ_ENUM_869
# define BOOST_PP_SEQ_ENUM_871(x) x, BOOST_PP_SEQ_ENUM_870
# define BOOST_PP_SEQ_ENUM_872(x) x, BOOST_PP_SEQ_ENUM_871
# define BOOST_PP_SEQ_ENUM_873(x) x, BOOST_PP_SEQ_ENUM_872
# define BOOST_PP_SEQ_ENUM_874(x) x, BOOST_PP_SEQ_ENUM_873
# define BOOST_PP_SEQ_ENUM_875(x) x, BOOST_PP_SEQ_ENUM_874
# define BOOST_PP_SEQ_ENUM_876(x) x, BOOST_PP_SEQ_ENUM_875
# define BOOST_PP_SEQ_ENUM_877(x) x, BOOST_PP_SEQ_ENUM_876
# define BOOST_PP_SEQ_ENUM_878(x) x, BOOST_PP_SEQ_ENUM_877
# define BOOST_PP_SEQ_ENUM_879(x) x, BOOST_PP_SEQ_ENUM_878
# define BOOST_PP_SEQ_ENUM_880(x) x, BOOST_PP_SEQ_ENUM_879
# define BOOST_PP_SEQ_ENUM_881(x) x, BOOST_PP_SEQ_ENUM_880
# define BOOST_PP_SEQ_ENUM_882(x) x, BOOST_PP_SEQ_ENUM_881
# define BOOST_PP_SEQ_ENUM_883(x) x, BOOST_PP_SEQ_ENUM_882
# define BOOST_PP_SEQ_ENUM_884(x) x, BOOST_PP_SEQ_ENUM_883
# define BOOST_PP_SEQ_ENUM_885(x) x, BOOST_PP_SEQ_ENUM_884
# define BOOST_PP_SEQ_ENUM_886(x) x, BOOST_PP_SEQ_ENUM_885
# define BOOST_PP_SEQ_ENUM_887(x) x, BOOST_PP_SEQ_ENUM_886
# define BOOST_PP_SEQ_ENUM_888(x) x, BOOST_PP_SEQ_ENUM_887
# define BOOST_PP_SEQ_ENUM_889(x) x, BOOST_PP_SEQ_ENUM_888
# define BOOST_PP_SEQ_ENUM_890(x) x, BOOST_PP_SEQ_ENUM_889
# define BOOST_PP_SEQ_ENUM_891(x) x, BOOST_PP_SEQ_ENUM_890
# define BOOST_PP_SEQ_ENUM_892(x) x, BOOST_PP_SEQ_ENUM_891
# define BOOST_PP_SEQ_ENUM_893(x) x, BOOST_PP_SEQ_ENUM_892
# define BOOST_PP_SEQ_ENUM_894(x) x, BOOST_PP_SEQ_ENUM_893
# define BOOST_PP_SEQ_ENUM_895(x) x, BOOST_PP_SEQ_ENUM_894
# define BOOST_PP_SEQ_ENUM_896(x) x, BOOST_PP_SEQ_ENUM_895
# define BOOST_PP_SEQ_ENUM_897(x) x, BOOST_PP_SEQ_ENUM_896
# define BOOST_PP_SEQ_ENUM_898(x) x, BOOST_PP_SEQ_ENUM_897
# define BOOST_PP_SEQ_ENUM_899(x) x, BOOST_PP_SEQ_ENUM_898
# define BOOST_PP_SEQ_ENUM_900(x) x, BOOST_PP_SEQ_ENUM_899
# define BOOST_PP_SEQ_ENUM_901(x) x, BOOST_PP_SEQ_ENUM_900
# define BOOST_PP_SEQ_ENUM_902(x) x, BOOST_PP_SEQ_ENUM_901
# define BOOST_PP_SEQ_ENUM_903(x) x, BOOST_PP_SEQ_ENUM_902
# define BOOST_PP_SEQ_ENUM_904(x) x, BOOST_PP_SEQ_ENUM_903
# define BOOST_PP_SEQ_ENUM_905(x) x, BOOST_PP_SEQ_ENUM_904
# define BOOST_PP_SEQ_ENUM_906(x) x, BOOST_PP_SEQ_ENUM_905
# define BOOST_PP_SEQ_ENUM_907(x) x, BOOST_PP_SEQ_ENUM_906
# define BOOST_PP_SEQ_ENUM_908(x) x, BOOST_PP_SEQ_ENUM_907
# define BOOST_PP_SEQ_ENUM_909(x) x, BOOST_PP_SEQ_ENUM_908
# define BOOST_PP_SEQ_ENUM_910(x) x, BOOST_PP_SEQ_ENUM_909
# define BOOST_PP_SEQ_ENUM_911(x) x, BOOST_PP_SEQ_ENUM_910
# define BOOST_PP_SEQ_ENUM_912(x) x, BOOST_PP_SEQ_ENUM_911
# define BOOST_PP_SEQ_ENUM_913(x) x, BOOST_PP_SEQ_ENUM_912
# define BOOST_PP_SEQ_ENUM_914(x) x, BOOST_PP_SEQ_ENUM_913
# define BOOST_PP_SEQ_ENUM_915(x) x, BOOST_PP_SEQ_ENUM_914
# define BOOST_PP_SEQ_ENUM_916(x) x, BOOST_PP_SEQ_ENUM_915
# define BOOST_PP_SEQ_ENUM_917(x) x, BOOST_PP_SEQ_ENUM_916
# define BOOST_PP_SEQ_ENUM_918(x) x, BOOST_PP_SEQ_ENUM_917
# define BOOST_PP_SEQ_ENUM_919(x) x, BOOST_PP_SEQ_ENUM_918
# define BOOST_PP_SEQ_ENUM_920(x) x, BOOST_PP_SEQ_ENUM_919
# define BOOST_PP_SEQ_ENUM_921(x) x, BOOST_PP_SEQ_ENUM_920
# define BOOST_PP_SEQ_ENUM_922(x) x, BOOST_PP_SEQ_ENUM_921
# define BOOST_PP_SEQ_ENUM_923(x) x, BOOST_PP_SEQ_ENUM_922
# define BOOST_PP_SEQ_ENUM_924(x) x, BOOST_PP_SEQ_ENUM_923
# define BOOST_PP_SEQ_ENUM_925(x) x, BOOST_PP_SEQ_ENUM_924
# define BOOST_PP_SEQ_ENUM_926(x) x, BOOST_PP_SEQ_ENUM_925
# define BOOST_PP_SEQ_ENUM_927(x) x, BOOST_PP_SEQ_ENUM_926
# define BOOST_PP_SEQ_ENUM_928(x) x, BOOST_PP_SEQ_ENUM_927
# define BOOST_PP_SEQ_ENUM_929(x) x, BOOST_PP_SEQ_ENUM_928
# define BOOST_PP_SEQ_ENUM_930(x) x, BOOST_PP_SEQ_ENUM_929
# define BOOST_PP_SEQ_ENUM_931(x) x, BOOST_PP_SEQ_ENUM_930
# define BOOST_PP_SEQ_ENUM_932(x) x, BOOST_PP_SEQ_ENUM_931
# define BOOST_PP_SEQ_ENUM_933(x) x, BOOST_PP_SEQ_ENUM_932
# define BOOST_PP_SEQ_ENUM_934(x) x, BOOST_PP_SEQ_ENUM_933
# define BOOST_PP_SEQ_ENUM_935(x) x, BOOST_PP_SEQ_ENUM_934
# define BOOST_PP_SEQ_ENUM_936(x) x, BOOST_PP_SEQ_ENUM_935
# define BOOST_PP_SEQ_ENUM_937(x) x, BOOST_PP_SEQ_ENUM_936
# define BOOST_PP_SEQ_ENUM_938(x) x, BOOST_PP_SEQ_ENUM_937
# define BOOST_PP_SEQ_ENUM_939(x) x, BOOST_PP_SEQ_ENUM_938
# define BOOST_PP_SEQ_ENUM_940(x) x, BOOST_PP_SEQ_ENUM_939
# define BOOST_PP_SEQ_ENUM_941(x) x, BOOST_PP_SEQ_ENUM_940
# define BOOST_PP_SEQ_ENUM_942(x) x, BOOST_PP_SEQ_ENUM_941
# define BOOST_PP_SEQ_ENUM_943(x) x, BOOST_PP_SEQ_ENUM_942
# define BOOST_PP_SEQ_ENUM_944(x) x, BOOST_PP_SEQ_ENUM_943
# define BOOST_PP_SEQ_ENUM_945(x) x, BOOST_PP_SEQ_ENUM_944
# define BOOST_PP_SEQ_ENUM_946(x) x, BOOST_PP_SEQ_ENUM_945
# define BOOST_PP_SEQ_ENUM_947(x) x, BOOST_PP_SEQ_ENUM_946
# define BOOST_PP_SEQ_ENUM_948(x) x, BOOST_PP_SEQ_ENUM_947
# define BOOST_PP_SEQ_ENUM_949(x) x, BOOST_PP_SEQ_ENUM_948
# define BOOST_PP_SEQ_ENUM_950(x) x, BOOST_PP_SEQ_ENUM_949
# define BOOST_PP_SEQ_ENUM_951(x) x, BOOST_PP_SEQ_ENUM_950
# define BOOST_PP_SEQ_ENUM_952(x) x, BOOST_PP_SEQ_ENUM_951
# define BOOST_PP_SEQ_ENUM_953(x) x, BOOST_PP_SEQ_ENUM_952
# define BOOST_PP_SEQ_ENUM_954(x) x, BOOST_PP_SEQ_ENUM_953
# define BOOST_PP_SEQ_ENUM_955(x) x, BOOST_PP_SEQ_ENUM_954
# define BOOST_PP_SEQ_ENUM_956(x) x, BOOST_PP_SEQ_ENUM_955
# define BOOST_PP_SEQ_ENUM_957(x) x, BOOST_PP_SEQ_ENUM_956
# define BOOST_PP_SEQ_ENUM_958(x) x, BOOST_PP_SEQ_ENUM_957
# define BOOST_PP_SEQ_ENUM_959(x) x, BOOST_PP_SEQ_ENUM_958
# define BOOST_PP_SEQ_ENUM_960(x) x, BOOST_PP_SEQ_ENUM_959
# define BOOST_PP_SEQ_ENUM_961(x) x, BOOST_PP_SEQ_ENUM_960
# define BOOST_PP_SEQ_ENUM_962(x) x, BOOST_PP_SEQ_ENUM_961
# define BOOST_PP_SEQ_ENUM_963(x) x, BOOST_PP_SEQ_ENUM_962
# define BOOST_PP_SEQ_ENUM_964(x) x, BOOST_PP_SEQ_ENUM_963
# define BOOST_PP_SEQ_ENUM_965(x) x, BOOST_PP_SEQ_ENUM_964
# define BOOST_PP_SEQ_ENUM_966(x) x, BOOST_PP_SEQ_ENUM_965
# define BOOST_PP_SEQ_ENUM_967(x) x, BOOST_PP_SEQ_ENUM_966
# define BOOST_PP_SEQ_ENUM_968(x) x, BOOST_PP_SEQ_ENUM_967
# define BOOST_PP_SEQ_ENUM_969(x) x, BOOST_PP_SEQ_ENUM_968
# define BOOST_PP_SEQ_ENUM_970(x) x, BOOST_PP_SEQ_ENUM_969
# define BOOST_PP_SEQ_ENUM_971(x) x, BOOST_PP_SEQ_ENUM_970
# define BOOST_PP_SEQ_ENUM_972(x) x, BOOST_PP_SEQ_ENUM_971
# define BOOST_PP_SEQ_ENUM_973(x) x, BOOST_PP_SEQ_ENUM_972
# define BOOST_PP_SEQ_ENUM_974(x) x, BOOST_PP_SEQ_ENUM_973
# define BOOST_PP_SEQ_ENUM_975(x) x, BOOST_PP_SEQ_ENUM_974
# define BOOST_PP_SEQ_ENUM_976(x) x, BOOST_PP_SEQ_ENUM_975
# define BOOST_PP_SEQ_ENUM_977(x) x, BOOST_PP_SEQ_ENUM_976
# define BOOST_PP_SEQ_ENUM_978(x) x, BOOST_PP_SEQ_ENUM_977
# define BOOST_PP_SEQ_ENUM_979(x) x, BOOST_PP_SEQ_ENUM_978
# define BOOST_PP_SEQ_ENUM_980(x) x, BOOST_PP_SEQ_ENUM_979
# define BOOST_PP_SEQ_ENUM_981(x) x, BOOST_PP_SEQ_ENUM_980
# define BOOST_PP_SEQ_ENUM_982(x) x, BOOST_PP_SEQ_ENUM_981
# define BOOST_PP_SEQ_ENUM_983(x) x, BOOST_PP_SEQ_ENUM_982
# define BOOST_PP_SEQ_ENUM_984(x) x, BOOST_PP_SEQ_ENUM_983
# define BOOST_PP_SEQ_ENUM_985(x) x, BOOST_PP_SEQ_ENUM_984
# define BOOST_PP_SEQ_ENUM_986(x) x, BOOST_PP_SEQ_ENUM_985
# define BOOST_PP_SEQ_ENUM_987(x) x, BOOST_PP_SEQ_ENUM_986
# define BOOST_PP_SEQ_ENUM_988(x) x, BOOST_PP_SEQ_ENUM_987
# define BOOST_PP_SEQ_ENUM_989(x) x, BOOST_PP_SEQ_ENUM_988
# define BOOST_PP_SEQ_ENUM_990(x) x, BOOST_PP_SEQ_ENUM_989
# define BOOST_PP_SEQ_ENUM_991(x) x, BOOST_PP_SEQ_ENUM_990
# define BOOST_PP_SEQ_ENUM_992(x) x, BOOST_PP_SEQ_ENUM_991
# define BOOST_PP_SEQ_ENUM_993(x) x, BOOST_PP_SEQ_ENUM_992
# define BOOST_PP_SEQ_ENUM_994(x) x, BOOST_PP_SEQ_ENUM_993
# define BOOST_PP_SEQ_ENUM_995(x) x, BOOST_PP_SEQ_ENUM_994
# define BOOST_PP_SEQ_ENUM_996(x) x, BOOST_PP_SEQ_ENUM_995
# define BOOST_PP_SEQ_ENUM_997(x) x, BOOST_PP_SEQ_ENUM_996
# define BOOST_PP_SEQ_ENUM_998(x) x, BOOST_PP_SEQ_ENUM_997
# define BOOST_PP_SEQ_ENUM_999(x) x, BOOST_PP_SEQ_ENUM_998
# define BOOST_PP_SEQ_ENUM_1000(x) x, BOOST_PP_SEQ_ENUM_999
# define BOOST_PP_SEQ_ENUM_1001(x) x, BOOST_PP_SEQ_ENUM_1000
# define BOOST_PP_SEQ_ENUM_1002(x) x, BOOST_PP_SEQ_ENUM_1001
# define BOOST_PP_SEQ_ENUM_1003(x) x, BOOST_PP_SEQ_ENUM_1002
# define BOOST_PP_SEQ_ENUM_1004(x) x, BOOST_PP_SEQ_ENUM_1003
# define BOOST_PP_SEQ_ENUM_1005(x) x, BOOST_PP_SEQ_ENUM_1004
# define BOOST_PP_SEQ_ENUM_1006(x) x, BOOST_PP_SEQ_ENUM_1005
# define BOOST_PP_SEQ_ENUM_1007(x) x, BOOST_PP_SEQ_ENUM_1006
# define BOOST_PP_SEQ_ENUM_1008(x) x, BOOST_PP_SEQ_ENUM_1007
# define BOOST_PP_SEQ_ENUM_1009(x) x, BOOST_PP_SEQ_ENUM_1008
# define BOOST_PP_SEQ_ENUM_1010(x) x, BOOST_PP_SEQ_ENUM_1009
# define BOOST_PP_SEQ_ENUM_1011(x) x, BOOST_PP_SEQ_ENUM_1010
# define BOOST_PP_SEQ_ENUM_1012(x) x, BOOST_PP_SEQ_ENUM_1011
# define BOOST_PP_SEQ_ENUM_1013(x) x, BOOST_PP_SEQ_ENUM_1012
# define BOOST_PP_SEQ_ENUM_1014(x) x, BOOST_PP_SEQ_ENUM_1013
# define BOOST_PP_SEQ_ENUM_1015(x) x, BOOST_PP_SEQ_ENUM_1014
# define BOOST_PP_SEQ_ENUM_1016(x) x, BOOST_PP_SEQ_ENUM_1015
# define BOOST_PP_SEQ_ENUM_1017(x) x, BOOST_PP_SEQ_ENUM_1016
# define BOOST_PP_SEQ_ENUM_1018(x) x, BOOST_PP_SEQ_ENUM_1017
# define BOOST_PP_SEQ_ENUM_1019(x) x, BOOST_PP_SEQ_ENUM_1018
# define BOOST_PP_SEQ_ENUM_1020(x) x, BOOST_PP_SEQ_ENUM_1019
# define BOOST_PP_SEQ_ENUM_1021(x) x, BOOST_PP_SEQ_ENUM_1020
# define BOOST_PP_SEQ_ENUM_1022(x) x, BOOST_PP_SEQ_ENUM_1021
# define BOOST_PP_SEQ_ENUM_1023(x) x, BOOST_PP_SEQ_ENUM_1022
# define BOOST_PP_SEQ_ENUM_1024(x) x, BOOST_PP_SEQ_ENUM_1023
#
# endif
