# /* Copyright (C) 2001
#  * Housemarque Oy
#  * http://www.housemarque.com
#  *
#  * Distributed under the Boost Software License, Version 1.0. (See
#  * accompanying file LICENSE_1_0.txt or copy at
#  * http://www.boost.org/LICENSE_1_0.txt)
#  */
#
# /* Revised by Paul Mensonides (2002) */
# /* Revised by Edward Diener (2020) */
#
#
# /* See http://www.boost.org for most recent version. */
#
# ifndef BOOST_PREPROCESSOR_LIST_DETAIL_FOLD_RIGHT_1024_HPP
# define BOOST_PREPROCESSOR_LIST_DETAIL_FOLD_RIGHT_1024_HPP
#
# define BOOST_PP_LIST_FOLD_RIGHT_513(o, s, l) BOOST_PP_LIST_FOLD_LEFT_513(o, s, BOOST_PP_LIST_REVERSE_D(513, l))
# define BOOST_PP_LIST_FOLD_RIGHT_514(o, s, l) BOOST_PP_LIST_FOLD_LEFT_514(o, s, BOOST_PP_LIST_REVERSE_D(514, l))
# define BOOST_PP_LIST_FOLD_RIGHT_515(o, s, l) BOOST_PP_LIST_FOLD_LEFT_515(o, s, BOOST_PP_LIST_REVERSE_D(515, l))
# define BOOST_PP_LIST_FOLD_RIGHT_516(o, s, l) BOOST_PP_LIST_FOLD_LEFT_516(o, s, BOOST_PP_LIST_REVERSE_D(516, l))
# define BOOST_PP_LIST_FOLD_RIGHT_517(o, s, l) BOOST_PP_LIST_FOLD_LEFT_517(o, s, BOOST_PP_LIST_REVERSE_D(517, l))
# define BOOST_PP_LIST_FOLD_RIGHT_518(o, s, l) BOOST_PP_LIST_FOLD_LEFT_518(o, s, BOOST_PP_LIST_REVERSE_D(518, l))
# define BOOST_PP_LIST_FOLD_RIGHT_519(o, s, l) BOOST_PP_LIST_FOLD_LEFT_519(o, s, BOOST_PP_LIST_REVERSE_D(519, l))
# define BOOST_PP_LIST_FOLD_RIGHT_520(o, s, l) BOOST_PP_LIST_FOLD_LEFT_520(o, s, BOOST_PP_LIST_REVERSE_D(520, l))
# define BOOST_PP_LIST_FOLD_RIGHT_521(o, s, l) BOOST_PP_LIST_FOLD_LEFT_521(o, s, BOOST_PP_LIST_REVERSE_D(521, l))
# define BOOST_PP_LIST_FOLD_RIGHT_522(o, s, l) BOOST_PP_LIST_FOLD_LEFT_522(o, s, BOOST_PP_LIST_REVERSE_D(522, l))
# define BOOST_PP_LIST_FOLD_RIGHT_523(o, s, l) BOOST_PP_LIST_FOLD_LEFT_523(o, s, BOOST_PP_LIST_REVERSE_D(523, l))
# define BOOST_PP_LIST_FOLD_RIGHT_524(o, s, l) BOOST_PP_LIST_FOLD_LEFT_524(o, s, BOOST_PP_LIST_REVERSE_D(524, l))
# define BOOST_PP_LIST_FOLD_RIGHT_525(o, s, l) BOOST_PP_LIST_FOLD_LEFT_525(o, s, BOOST_PP_LIST_REVERSE_D(525, l))
# define BOOST_PP_LIST_FOLD_RIGHT_526(o, s, l) BOOST_PP_LIST_FOLD_LEFT_526(o, s, BOOST_PP_LIST_REVERSE_D(526, l))
# define BOOST_PP_LIST_FOLD_RIGHT_527(o, s, l) BOOST_PP_LIST_FOLD_LEFT_527(o, s, BOOST_PP_LIST_REVERSE_D(527, l))
# define BOOST_PP_LIST_FOLD_RIGHT_528(o, s, l) BOOST_PP_LIST_FOLD_LEFT_528(o, s, BOOST_PP_LIST_REVERSE_D(528, l))
# define BOOST_PP_LIST_FOLD_RIGHT_529(o, s, l) BOOST_PP_LIST_FOLD_LEFT_529(o, s, BOOST_PP_LIST_REVERSE_D(529, l))
# define BOOST_PP_LIST_FOLD_RIGHT_530(o, s, l) BOOST_PP_LIST_FOLD_LEFT_530(o, s, BOOST_PP_LIST_REVERSE_D(530, l))
# define BOOST_PP_LIST_FOLD_RIGHT_531(o, s, l) BOOST_PP_LIST_FOLD_LEFT_531(o, s, BOOST_PP_LIST_REVERSE_D(531, l))
# define BOOST_PP_LIST_FOLD_RIGHT_532(o, s, l) BOOST_PP_LIST_FOLD_LEFT_532(o, s, BOOST_PP_LIST_REVERSE_D(532, l))
# define BOOST_PP_LIST_FOLD_RIGHT_533(o, s, l) BOOST_PP_LIST_FOLD_LEFT_533(o, s, BOOST_PP_LIST_REVERSE_D(533, l))
# define BOOST_PP_LIST_FOLD_RIGHT_534(o, s, l) BOOST_PP_LIST_FOLD_LEFT_534(o, s, BOOST_PP_LIST_REVERSE_D(534, l))
# define BOOST_PP_LIST_FOLD_RIGHT_535(o, s, l) BOOST_PP_LIST_FOLD_LEFT_535(o, s, BOOST_PP_LIST_REVERSE_D(535, l))
# define BOOST_PP_LIST_FOLD_RIGHT_536(o, s, l) BOOST_PP_LIST_FOLD_LEFT_536(o, s, BOOST_PP_LIST_REVERSE_D(536, l))
# define BOOST_PP_LIST_FOLD_RIGHT_537(o, s, l) BOOST_PP_LIST_FOLD_LEFT_537(o, s, BOOST_PP_LIST_REVERSE_D(537, l))
# define BOOST_PP_LIST_FOLD_RIGHT_538(o, s, l) BOOST_PP_LIST_FOLD_LEFT_538(o, s, BOOST_PP_LIST_REVERSE_D(538, l))
# define BOOST_PP_LIST_FOLD_RIGHT_539(o, s, l) BOOST_PP_LIST_FOLD_LEFT_539(o, s, BOOST_PP_LIST_REVERSE_D(539, l))
# define BOOST_PP_LIST_FOLD_RIGHT_540(o, s, l) BOOST_PP_LIST_FOLD_LEFT_540(o, s, BOOST_PP_LIST_REVERSE_D(540, l))
# define BOOST_PP_LIST_FOLD_RIGHT_541(o, s, l) BOOST_PP_LIST_FOLD_LEFT_541(o, s, BOOST_PP_LIST_REVERSE_D(541, l))
# define BOOST_PP_LIST_FOLD_RIGHT_542(o, s, l) BOOST_PP_LIST_FOLD_LEFT_542(o, s, BOOST_PP_LIST_REVERSE_D(542, l))
# define BOOST_PP_LIST_FOLD_RIGHT_543(o, s, l) BOOST_PP_LIST_FOLD_LEFT_543(o, s, BOOST_PP_LIST_REVERSE_D(543, l))
# define BOOST_PP_LIST_FOLD_RIGHT_544(o, s, l) BOOST_PP_LIST_FOLD_LEFT_544(o, s, BOOST_PP_LIST_REVERSE_D(544, l))
# define BOOST_PP_LIST_FOLD_RIGHT_545(o, s, l) BOOST_PP_LIST_FOLD_LEFT_545(o, s, BOOST_PP_LIST_REVERSE_D(545, l))
# define BOOST_PP_LIST_FOLD_RIGHT_546(o, s, l) BOOST_PP_LIST_FOLD_LEFT_546(o, s, BOOST_PP_LIST_REVERSE_D(546, l))
# define BOOST_PP_LIST_FOLD_RIGHT_547(o, s, l) BOOST_PP_LIST_FOLD_LEFT_547(o, s, BOOST_PP_LIST_REVERSE_D(547, l))
# define BOOST_PP_LIST_FOLD_RIGHT_548(o, s, l) BOOST_PP_LIST_FOLD_LEFT_548(o, s, BOOST_PP_LIST_REVERSE_D(548, l))
# define BOOST_PP_LIST_FOLD_RIGHT_549(o, s, l) BOOST_PP_LIST_FOLD_LEFT_549(o, s, BOOST_PP_LIST_REVERSE_D(549, l))
# define BOOST_PP_LIST_FOLD_RIGHT_550(o, s, l) BOOST_PP_LIST_FOLD_LEFT_550(o, s, BOOST_PP_LIST_REVERSE_D(550, l))
# define BOOST_PP_LIST_FOLD_RIGHT_551(o, s, l) BOOST_PP_LIST_FOLD_LEFT_551(o, s, BOOST_PP_LIST_REVERSE_D(551, l))
# define BOOST_PP_LIST_FOLD_RIGHT_552(o, s, l) BOOST_PP_LIST_FOLD_LEFT_552(o, s, BOOST_PP_LIST_REVERSE_D(552, l))
# define BOOST_PP_LIST_FOLD_RIGHT_553(o, s, l) BOOST_PP_LIST_FOLD_LEFT_553(o, s, BOOST_PP_LIST_REVERSE_D(553, l))
# define BOOST_PP_LIST_FOLD_RIGHT_554(o, s, l) BOOST_PP_LIST_FOLD_LEFT_554(o, s, BOOST_PP_LIST_REVERSE_D(554, l))
# define BOOST_PP_LIST_FOLD_RIGHT_555(o, s, l) BOOST_PP_LIST_FOLD_LEFT_555(o, s, BOOST_PP_LIST_REVERSE_D(555, l))
# define BOOST_PP_LIST_FOLD_RIGHT_556(o, s, l) BOOST_PP_LIST_FOLD_LEFT_556(o, s, BOOST_PP_LIST_REVERSE_D(556, l))
# define BOOST_PP_LIST_FOLD_RIGHT_557(o, s, l) BOOST_PP_LIST_FOLD_LEFT_557(o, s, BOOST_PP_LIST_REVERSE_D(557, l))
# define BOOST_PP_LIST_FOLD_RIGHT_558(o, s, l) BOOST_PP_LIST_FOLD_LEFT_558(o, s, BOOST_PP_LIST_REVERSE_D(558, l))
# define BOOST_PP_LIST_FOLD_RIGHT_559(o, s, l) BOOST_PP_LIST_FOLD_LEFT_559(o, s, BOOST_PP_LIST_REVERSE_D(559, l))
# define BOOST_PP_LIST_FOLD_RIGHT_560(o, s, l) BOOST_PP_LIST_FOLD_LEFT_560(o, s, BOOST_PP_LIST_REVERSE_D(560, l))
# define BOOST_PP_LIST_FOLD_RIGHT_561(o, s, l) BOOST_PP_LIST_FOLD_LEFT_561(o, s, BOOST_PP_LIST_REVERSE_D(561, l))
# define BOOST_PP_LIST_FOLD_RIGHT_562(o, s, l) BOOST_PP_LIST_FOLD_LEFT_562(o, s, BOOST_PP_LIST_REVERSE_D(562, l))
# define BOOST_PP_LIST_FOLD_RIGHT_563(o, s, l) BOOST_PP_LIST_FOLD_LEFT_563(o, s, BOOST_PP_LIST_REVERSE_D(563, l))
# define BOOST_PP_LIST_FOLD_RIGHT_564(o, s, l) BOOST_PP_LIST_FOLD_LEFT_564(o, s, BOOST_PP_LIST_REVERSE_D(564, l))
# define BOOST_PP_LIST_FOLD_RIGHT_565(o, s, l) BOOST_PP_LIST_FOLD_LEFT_565(o, s, BOOST_PP_LIST_REVERSE_D(565, l))
# define BOOST_PP_LIST_FOLD_RIGHT_566(o, s, l) BOOST_PP_LIST_FOLD_LEFT_566(o, s, BOOST_PP_LIST_REVERSE_D(566, l))
# define BOOST_PP_LIST_FOLD_RIGHT_567(o, s, l) BOOST_PP_LIST_FOLD_LEFT_567(o, s, BOOST_PP_LIST_REVERSE_D(567, l))
# define BOOST_PP_LIST_FOLD_RIGHT_568(o, s, l) BOOST_PP_LIST_FOLD_LEFT_568(o, s, BOOST_PP_LIST_REVERSE_D(568, l))
# define BOOST_PP_LIST_FOLD_RIGHT_569(o, s, l) BOOST_PP_LIST_FOLD_LEFT_569(o, s, BOOST_PP_LIST_REVERSE_D(569, l))
# define BOOST_PP_LIST_FOLD_RIGHT_570(o, s, l) BOOST_PP_LIST_FOLD_LEFT_570(o, s, BOOST_PP_LIST_REVERSE_D(570, l))
# define BOOST_PP_LIST_FOLD_RIGHT_571(o, s, l) BOOST_PP_LIST_FOLD_LEFT_571(o, s, BOOST_PP_LIST_REVERSE_D(571, l))
# define BOOST_PP_LIST_FOLD_RIGHT_572(o, s, l) BOOST_PP_LIST_FOLD_LEFT_572(o, s, BOOST_PP_LIST_REVERSE_D(572, l))
# define BOOST_PP_LIST_FOLD_RIGHT_573(o, s, l) BOOST_PP_LIST_FOLD_LEFT_573(o, s, BOOST_PP_LIST_REVERSE_D(573, l))
# define BOOST_PP_LIST_FOLD_RIGHT_574(o, s, l) BOOST_PP_LIST_FOLD_LEFT_574(o, s, BOOST_PP_LIST_REVERSE_D(574, l))
# define BOOST_PP_LIST_FOLD_RIGHT_575(o, s, l) BOOST_PP_LIST_FOLD_LEFT_575(o, s, BOOST_PP_LIST_REVERSE_D(575, l))
# define BOOST_PP_LIST_FOLD_RIGHT_576(o, s, l) BOOST_PP_LIST_FOLD_LEFT_576(o, s, BOOST_PP_LIST_REVERSE_D(576, l))
# define BOOST_PP_LIST_FOLD_RIGHT_577(o, s, l) BOOST_PP_LIST_FOLD_LEFT_577(o, s, BOOST_PP_LIST_REVERSE_D(577, l))
# define BOOST_PP_LIST_FOLD_RIGHT_578(o, s, l) BOOST_PP_LIST_FOLD_LEFT_578(o, s, BOOST_PP_LIST_REVERSE_D(578, l))
# define BOOST_PP_LIST_FOLD_RIGHT_579(o, s, l) BOOST_PP_LIST_FOLD_LEFT_579(o, s, BOOST_PP_LIST_REVERSE_D(579, l))
# define BOOST_PP_LIST_FOLD_RIGHT_580(o, s, l) BOOST_PP_LIST_FOLD_LEFT_580(o, s, BOOST_PP_LIST_REVERSE_D(580, l))
# define BOOST_PP_LIST_FOLD_RIGHT_581(o, s, l) BOOST_PP_LIST_FOLD_LEFT_581(o, s, BOOST_PP_LIST_REVERSE_D(581, l))
# define BOOST_PP_LIST_FOLD_RIGHT_582(o, s, l) BOOST_PP_LIST_FOLD_LEFT_582(o, s, BOOST_PP_LIST_REVERSE_D(582, l))
# define BOOST_PP_LIST_FOLD_RIGHT_583(o, s, l) BOOST_PP_LIST_FOLD_LEFT_583(o, s, BOOST_PP_LIST_REVERSE_D(583, l))
# define BOOST_PP_LIST_FOLD_RIGHT_584(o, s, l) BOOST_PP_LIST_FOLD_LEFT_584(o, s, BOOST_PP_LIST_REVERSE_D(584, l))
# define BOOST_PP_LIST_FOLD_RIGHT_585(o, s, l) BOOST_PP_LIST_FOLD_LEFT_585(o, s, BOOST_PP_LIST_REVERSE_D(585, l))
# define BOOST_PP_LIST_FOLD_RIGHT_586(o, s, l) BOOST_PP_LIST_FOLD_LEFT_586(o, s, BOOST_PP_LIST_REVERSE_D(586, l))
# define BOOST_PP_LIST_FOLD_RIGHT_587(o, s, l) BOOST_PP_LIST_FOLD_LEFT_587(o, s, BOOST_PP_LIST_REVERSE_D(587, l))
# define BOOST_PP_LIST_FOLD_RIGHT_588(o, s, l) BOOST_PP_LIST_FOLD_LEFT_588(o, s, BOOST_PP_LIST_REVERSE_D(588, l))
# define BOOST_PP_LIST_FOLD_RIGHT_589(o, s, l) BOOST_PP_LIST_FOLD_LEFT_589(o, s, BOOST_PP_LIST_REVERSE_D(589, l))
# define BOOST_PP_LIST_FOLD_RIGHT_590(o, s, l) BOOST_PP_LIST_FOLD_LEFT_590(o, s, BOOST_PP_LIST_REVERSE_D(590, l))
# define BOOST_PP_LIST_FOLD_RIGHT_591(o, s, l) BOOST_PP_LIST_FOLD_LEFT_591(o, s, BOOST_PP_LIST_REVERSE_D(591, l))
# define BOOST_PP_LIST_FOLD_RIGHT_592(o, s, l) BOOST_PP_LIST_FOLD_LEFT_592(o, s, BOOST_PP_LIST_REVERSE_D(592, l))
# define BOOST_PP_LIST_FOLD_RIGHT_593(o, s, l) BOOST_PP_LIST_FOLD_LEFT_593(o, s, BOOST_PP_LIST_REVERSE_D(593, l))
# define BOOST_PP_LIST_FOLD_RIGHT_594(o, s, l) BOOST_PP_LIST_FOLD_LEFT_594(o, s, BOOST_PP_LIST_REVERSE_D(594, l))
# define BOOST_PP_LIST_FOLD_RIGHT_595(o, s, l) BOOST_PP_LIST_FOLD_LEFT_595(o, s, BOOST_PP_LIST_REVERSE_D(595, l))
# define BOOST_PP_LIST_FOLD_RIGHT_596(o, s, l) BOOST_PP_LIST_FOLD_LEFT_596(o, s, BOOST_PP_LIST_REVERSE_D(596, l))
# define BOOST_PP_LIST_FOLD_RIGHT_597(o, s, l) BOOST_PP_LIST_FOLD_LEFT_597(o, s, BOOST_PP_LIST_REVERSE_D(597, l))
# define BOOST_PP_LIST_FOLD_RIGHT_598(o, s, l) BOOST_PP_LIST_FOLD_LEFT_598(o, s, BOOST_PP_LIST_REVERSE_D(598, l))
# define BOOST_PP_LIST_FOLD_RIGHT_599(o, s, l) BOOST_PP_LIST_FOLD_LEFT_599(o, s, BOOST_PP_LIST_REVERSE_D(599, l))
# define BOOST_PP_LIST_FOLD_RIGHT_600(o, s, l) BOOST_PP_LIST_FOLD_LEFT_600(o, s, BOOST_PP_LIST_REVERSE_D(600, l))
# define BOOST_PP_LIST_FOLD_RIGHT_601(o, s, l) BOOST_PP_LIST_FOLD_LEFT_601(o, s, BOOST_PP_LIST_REVERSE_D(601, l))
# define BOOST_PP_LIST_FOLD_RIGHT_602(o, s, l) BOOST_PP_LIST_FOLD_LEFT_602(o, s, BOOST_PP_LIST_REVERSE_D(602, l))
# define BOOST_PP_LIST_FOLD_RIGHT_603(o, s, l) BOOST_PP_LIST_FOLD_LEFT_603(o, s, BOOST_PP_LIST_REVERSE_D(603, l))
# define BOOST_PP_LIST_FOLD_RIGHT_604(o, s, l) BOOST_PP_LIST_FOLD_LEFT_604(o, s, BOOST_PP_LIST_REVERSE_D(604, l))
# define BOOST_PP_LIST_FOLD_RIGHT_605(o, s, l) BOOST_PP_LIST_FOLD_LEFT_605(o, s, BOOST_PP_LIST_REVERSE_D(605, l))
# define BOOST_PP_LIST_FOLD_RIGHT_606(o, s, l) BOOST_PP_LIST_FOLD_LEFT_606(o, s, BOOST_PP_LIST_REVERSE_D(606, l))
# define BOOST_PP_LIST_FOLD_RIGHT_607(o, s, l) BOOST_PP_LIST_FOLD_LEFT_607(o, s, BOOST_PP_LIST_REVERSE_D(607, l))
# define BOOST_PP_LIST_FOLD_RIGHT_608(o, s, l) BOOST_PP_LIST_FOLD_LEFT_608(o, s, BOOST_PP_LIST_REVERSE_D(608, l))
# define BOOST_PP_LIST_FOLD_RIGHT_609(o, s, l) BOOST_PP_LIST_FOLD_LEFT_609(o, s, BOOST_PP_LIST_REVERSE_D(609, l))
# define BOOST_PP_LIST_FOLD_RIGHT_610(o, s, l) BOOST_PP_LIST_FOLD_LEFT_610(o, s, BOOST_PP_LIST_REVERSE_D(610, l))
# define BOOST_PP_LIST_FOLD_RIGHT_611(o, s, l) BOOST_PP_LIST_FOLD_LEFT_611(o, s, BOOST_PP_LIST_REVERSE_D(611, l))
# define BOOST_PP_LIST_FOLD_RIGHT_612(o, s, l) BOOST_PP_LIST_FOLD_LEFT_612(o, s, BOOST_PP_LIST_REVERSE_D(612, l))
# define BOOST_PP_LIST_FOLD_RIGHT_613(o, s, l) BOOST_PP_LIST_FOLD_LEFT_613(o, s, BOOST_PP_LIST_REVERSE_D(613, l))
# define BOOST_PP_LIST_FOLD_RIGHT_614(o, s, l) BOOST_PP_LIST_FOLD_LEFT_614(o, s, BOOST_PP_LIST_REVERSE_D(614, l))
# define BOOST_PP_LIST_FOLD_RIGHT_615(o, s, l) BOOST_PP_LIST_FOLD_LEFT_615(o, s, BOOST_PP_LIST_REVERSE_D(615, l))
# define BOOST_PP_LIST_FOLD_RIGHT_616(o, s, l) BOOST_PP_LIST_FOLD_LEFT_616(o, s, BOOST_PP_LIST_REVERSE_D(616, l))
# define BOOST_PP_LIST_FOLD_RIGHT_617(o, s, l) BOOST_PP_LIST_FOLD_LEFT_617(o, s, BOOST_PP_LIST_REVERSE_D(617, l))
# define BOOST_PP_LIST_FOLD_RIGHT_618(o, s, l) BOOST_PP_LIST_FOLD_LEFT_618(o, s, BOOST_PP_LIST_REVERSE_D(618, l))
# define BOOST_PP_LIST_FOLD_RIGHT_619(o, s, l) BOOST_PP_LIST_FOLD_LEFT_619(o, s, BOOST_PP_LIST_REVERSE_D(619, l))
# define BOOST_PP_LIST_FOLD_RIGHT_620(o, s, l) BOOST_PP_LIST_FOLD_LEFT_620(o, s, BOOST_PP_LIST_REVERSE_D(620, l))
# define BOOST_PP_LIST_FOLD_RIGHT_621(o, s, l) BOOST_PP_LIST_FOLD_LEFT_621(o, s, BOOST_PP_LIST_REVERSE_D(621, l))
# define BOOST_PP_LIST_FOLD_RIGHT_622(o, s, l) BOOST_PP_LIST_FOLD_LEFT_622(o, s, BOOST_PP_LIST_REVERSE_D(622, l))
# define BOOST_PP_LIST_FOLD_RIGHT_623(o, s, l) BOOST_PP_LIST_FOLD_LEFT_623(o, s, BOOST_PP_LIST_REVERSE_D(623, l))
# define BOOST_PP_LIST_FOLD_RIGHT_624(o, s, l) BOOST_PP_LIST_FOLD_LEFT_624(o, s, BOOST_PP_LIST_REVERSE_D(624, l))
# define BOOST_PP_LIST_FOLD_RIGHT_625(o, s, l) BOOST_PP_LIST_FOLD_LEFT_625(o, s, BOOST_PP_LIST_REVERSE_D(625, l))
# define BOOST_PP_LIST_FOLD_RIGHT_626(o, s, l) BOOST_PP_LIST_FOLD_LEFT_626(o, s, BOOST_PP_LIST_REVERSE_D(626, l))
# define BOOST_PP_LIST_FOLD_RIGHT_627(o, s, l) BOOST_PP_LIST_FOLD_LEFT_627(o, s, BOOST_PP_LIST_REVERSE_D(627, l))
# define BOOST_PP_LIST_FOLD_RIGHT_628(o, s, l) BOOST_PP_LIST_FOLD_LEFT_628(o, s, BOOST_PP_LIST_REVERSE_D(628, l))
# define BOOST_PP_LIST_FOLD_RIGHT_629(o, s, l) BOOST_PP_LIST_FOLD_LEFT_629(o, s, BOOST_PP_LIST_REVERSE_D(629, l))
# define BOOST_PP_LIST_FOLD_RIGHT_630(o, s, l) BOOST_PP_LIST_FOLD_LEFT_630(o, s, BOOST_PP_LIST_REVERSE_D(630, l))
# define BOOST_PP_LIST_FOLD_RIGHT_631(o, s, l) BOOST_PP_LIST_FOLD_LEFT_631(o, s, BOOST_PP_LIST_REVERSE_D(631, l))
# define BOOST_PP_LIST_FOLD_RIGHT_632(o, s, l) BOOST_PP_LIST_FOLD_LEFT_632(o, s, BOOST_PP_LIST_REVERSE_D(632, l))
# define BOOST_PP_LIST_FOLD_RIGHT_633(o, s, l) BOOST_PP_LIST_FOLD_LEFT_633(o, s, BOOST_PP_LIST_REVERSE_D(633, l))
# define BOOST_PP_LIST_FOLD_RIGHT_634(o, s, l) BOOST_PP_LIST_FOLD_LEFT_634(o, s, BOOST_PP_LIST_REVERSE_D(634, l))
# define BOOST_PP_LIST_FOLD_RIGHT_635(o, s, l) BOOST_PP_LIST_FOLD_LEFT_635(o, s, BOOST_PP_LIST_REVERSE_D(635, l))
# define BOOST_PP_LIST_FOLD_RIGHT_636(o, s, l) BOOST_PP_LIST_FOLD_LEFT_636(o, s, BOOST_PP_LIST_REVERSE_D(636, l))
# define BOOST_PP_LIST_FOLD_RIGHT_637(o, s, l) BOOST_PP_LIST_FOLD_LEFT_637(o, s, BOOST_PP_LIST_REVERSE_D(637, l))
# define BOOST_PP_LIST_FOLD_RIGHT_638(o, s, l) BOOST_PP_LIST_FOLD_LEFT_638(o, s, BOOST_PP_LIST_REVERSE_D(638, l))
# define BOOST_PP_LIST_FOLD_RIGHT_639(o, s, l) BOOST_PP_LIST_FOLD_LEFT_639(o, s, BOOST_PP_LIST_REVERSE_D(639, l))
# define BOOST_PP_LIST_FOLD_RIGHT_640(o, s, l) BOOST_PP_LIST_FOLD_LEFT_640(o, s, BOOST_PP_LIST_REVERSE_D(640, l))
# define BOOST_PP_LIST_FOLD_RIGHT_641(o, s, l) BOOST_PP_LIST_FOLD_LEFT_641(o, s, BOOST_PP_LIST_REVERSE_D(641, l))
# define BOOST_PP_LIST_FOLD_RIGHT_642(o, s, l) BOOST_PP_LIST_FOLD_LEFT_642(o, s, BOOST_PP_LIST_REVERSE_D(642, l))
# define BOOST_PP_LIST_FOLD_RIGHT_643(o, s, l) BOOST_PP_LIST_FOLD_LEFT_643(o, s, BOOST_PP_LIST_REVERSE_D(643, l))
# define BOOST_PP_LIST_FOLD_RIGHT_644(o, s, l) BOOST_PP_LIST_FOLD_LEFT_644(o, s, BOOST_PP_LIST_REVERSE_D(644, l))
# define BOOST_PP_LIST_FOLD_RIGHT_645(o, s, l) BOOST_PP_LIST_FOLD_LEFT_645(o, s, BOOST_PP_LIST_REVERSE_D(645, l))
# define BOOST_PP_LIST_FOLD_RIGHT_646(o, s, l) BOOST_PP_LIST_FOLD_LEFT_646(o, s, BOOST_PP_LIST_REVERSE_D(646, l))
# define BOOST_PP_LIST_FOLD_RIGHT_647(o, s, l) BOOST_PP_LIST_FOLD_LEFT_647(o, s, BOOST_PP_LIST_REVERSE_D(647, l))
# define BOOST_PP_LIST_FOLD_RIGHT_648(o, s, l) BOOST_PP_LIST_FOLD_LEFT_648(o, s, BOOST_PP_LIST_REVERSE_D(648, l))
# define BOOST_PP_LIST_FOLD_RIGHT_649(o, s, l) BOOST_PP_LIST_FOLD_LEFT_649(o, s, BOOST_PP_LIST_REVERSE_D(649, l))
# define BOOST_PP_LIST_FOLD_RIGHT_650(o, s, l) BOOST_PP_LIST_FOLD_LEFT_650(o, s, BOOST_PP_LIST_REVERSE_D(650, l))
# define BOOST_PP_LIST_FOLD_RIGHT_651(o, s, l) BOOST_PP_LIST_FOLD_LEFT_651(o, s, BOOST_PP_LIST_REVERSE_D(651, l))
# define BOOST_PP_LIST_FOLD_RIGHT_652(o, s, l) BOOST_PP_LIST_FOLD_LEFT_652(o, s, BOOST_PP_LIST_REVERSE_D(652, l))
# define BOOST_PP_LIST_FOLD_RIGHT_653(o, s, l) BOOST_PP_LIST_FOLD_LEFT_653(o, s, BOOST_PP_LIST_REVERSE_D(653, l))
# define BOOST_PP_LIST_FOLD_RIGHT_654(o, s, l) BOOST_PP_LIST_FOLD_LEFT_654(o, s, BOOST_PP_LIST_REVERSE_D(654, l))
# define BOOST_PP_LIST_FOLD_RIGHT_655(o, s, l) BOOST_PP_LIST_FOLD_LEFT_655(o, s, BOOST_PP_LIST_REVERSE_D(655, l))
# define BOOST_PP_LIST_FOLD_RIGHT_656(o, s, l) BOOST_PP_LIST_FOLD_LEFT_656(o, s, BOOST_PP_LIST_REVERSE_D(656, l))
# define BOOST_PP_LIST_FOLD_RIGHT_657(o, s, l) BOOST_PP_LIST_FOLD_LEFT_657(o, s, BOOST_PP_LIST_REVERSE_D(657, l))
# define BOOST_PP_LIST_FOLD_RIGHT_658(o, s, l) BOOST_PP_LIST_FOLD_LEFT_658(o, s, BOOST_PP_LIST_REVERSE_D(658, l))
# define BOOST_PP_LIST_FOLD_RIGHT_659(o, s, l) BOOST_PP_LIST_FOLD_LEFT_659(o, s, BOOST_PP_LIST_REVERSE_D(659, l))
# define BOOST_PP_LIST_FOLD_RIGHT_660(o, s, l) BOOST_PP_LIST_FOLD_LEFT_660(o, s, BOOST_PP_LIST_REVERSE_D(660, l))
# define BOOST_PP_LIST_FOLD_RIGHT_661(o, s, l) BOOST_PP_LIST_FOLD_LEFT_661(o, s, BOOST_PP_LIST_REVERSE_D(661, l))
# define BOOST_PP_LIST_FOLD_RIGHT_662(o, s, l) BOOST_PP_LIST_FOLD_LEFT_662(o, s, BOOST_PP_LIST_REVERSE_D(662, l))
# define BOOST_PP_LIST_FOLD_RIGHT_663(o, s, l) BOOST_PP_LIST_FOLD_LEFT_663(o, s, BOOST_PP_LIST_REVERSE_D(663, l))
# define BOOST_PP_LIST_FOLD_RIGHT_664(o, s, l) BOOST_PP_LIST_FOLD_LEFT_664(o, s, BOOST_PP_LIST_REVERSE_D(664, l))
# define BOOST_PP_LIST_FOLD_RIGHT_665(o, s, l) BOOST_PP_LIST_FOLD_LEFT_665(o, s, BOOST_PP_LIST_REVERSE_D(665, l))
# define BOOST_PP_LIST_FOLD_RIGHT_666(o, s, l) BOOST_PP_LIST_FOLD_LEFT_666(o, s, BOOST_PP_LIST_REVERSE_D(666, l))
# define BOOST_PP_LIST_FOLD_RIGHT_667(o, s, l) BOOST_PP_LIST_FOLD_LEFT_667(o, s, BOOST_PP_LIST_REVERSE_D(667, l))
# define BOOST_PP_LIST_FOLD_RIGHT_668(o, s, l) BOOST_PP_LIST_FOLD_LEFT_668(o, s, BOOST_PP_LIST_REVERSE_D(668, l))
# define BOOST_PP_LIST_FOLD_RIGHT_669(o, s, l) BOOST_PP_LIST_FOLD_LEFT_669(o, s, BOOST_PP_LIST_REVERSE_D(669, l))
# define BOOST_PP_LIST_FOLD_RIGHT_670(o, s, l) BOOST_PP_LIST_FOLD_LEFT_670(o, s, BOOST_PP_LIST_REVERSE_D(670, l))
# define BOOST_PP_LIST_FOLD_RIGHT_671(o, s, l) BOOST_PP_LIST_FOLD_LEFT_671(o, s, BOOST_PP_LIST_REVERSE_D(671, l))
# define BOOST_PP_LIST_FOLD_RIGHT_672(o, s, l) BOOST_PP_LIST_FOLD_LEFT_672(o, s, BOOST_PP_LIST_REVERSE_D(672, l))
# define BOOST_PP_LIST_FOLD_RIGHT_673(o, s, l) BOOST_PP_LIST_FOLD_LEFT_673(o, s, BOOST_PP_LIST_REVERSE_D(673, l))
# define BOOST_PP_LIST_FOLD_RIGHT_674(o, s, l) BOOST_PP_LIST_FOLD_LEFT_674(o, s, BOOST_PP_LIST_REVERSE_D(674, l))
# define BOOST_PP_LIST_FOLD_RIGHT_675(o, s, l) BOOST_PP_LIST_FOLD_LEFT_675(o, s, BOOST_PP_LIST_REVERSE_D(675, l))
# define BOOST_PP_LIST_FOLD_RIGHT_676(o, s, l) BOOST_PP_LIST_FOLD_LEFT_676(o, s, BOOST_PP_LIST_REVERSE_D(676, l))
# define BOOST_PP_LIST_FOLD_RIGHT_677(o, s, l) BOOST_PP_LIST_FOLD_LEFT_677(o, s, BOOST_PP_LIST_REVERSE_D(677, l))
# define BOOST_PP_LIST_FOLD_RIGHT_678(o, s, l) BOOST_PP_LIST_FOLD_LEFT_678(o, s, BOOST_PP_LIST_REVERSE_D(678, l))
# define BOOST_PP_LIST_FOLD_RIGHT_679(o, s, l) BOOST_PP_LIST_FOLD_LEFT_679(o, s, BOOST_PP_LIST_REVERSE_D(679, l))
# define BOOST_PP_LIST_FOLD_RIGHT_680(o, s, l) BOOST_PP_LIST_FOLD_LEFT_680(o, s, BOOST_PP_LIST_REVERSE_D(680, l))
# define BOOST_PP_LIST_FOLD_RIGHT_681(o, s, l) BOOST_PP_LIST_FOLD_LEFT_681(o, s, BOOST_PP_LIST_REVERSE_D(681, l))
# define BOOST_PP_LIST_FOLD_RIGHT_682(o, s, l) BOOST_PP_LIST_FOLD_LEFT_682(o, s, BOOST_PP_LIST_REVERSE_D(682, l))
# define BOOST_PP_LIST_FOLD_RIGHT_683(o, s, l) BOOST_PP_LIST_FOLD_LEFT_683(o, s, BOOST_PP_LIST_REVERSE_D(683, l))
# define BOOST_PP_LIST_FOLD_RIGHT_684(o, s, l) BOOST_PP_LIST_FOLD_LEFT_684(o, s, BOOST_PP_LIST_REVERSE_D(684, l))
# define BOOST_PP_LIST_FOLD_RIGHT_685(o, s, l) BOOST_PP_LIST_FOLD_LEFT_685(o, s, BOOST_PP_LIST_REVERSE_D(685, l))
# define BOOST_PP_LIST_FOLD_RIGHT_686(o, s, l) BOOST_PP_LIST_FOLD_LEFT_686(o, s, BOOST_PP_LIST_REVERSE_D(686, l))
# define BOOST_PP_LIST_FOLD_RIGHT_687(o, s, l) BOOST_PP_LIST_FOLD_LEFT_687(o, s, BOOST_PP_LIST_REVERSE_D(687, l))
# define BOOST_PP_LIST_FOLD_RIGHT_688(o, s, l) BOOST_PP_LIST_FOLD_LEFT_688(o, s, BOOST_PP_LIST_REVERSE_D(688, l))
# define BOOST_PP_LIST_FOLD_RIGHT_689(o, s, l) BOOST_PP_LIST_FOLD_LEFT_689(o, s, BOOST_PP_LIST_REVERSE_D(689, l))
# define BOOST_PP_LIST_FOLD_RIGHT_690(o, s, l) BOOST_PP_LIST_FOLD_LEFT_690(o, s, BOOST_PP_LIST_REVERSE_D(690, l))
# define BOOST_PP_LIST_FOLD_RIGHT_691(o, s, l) BOOST_PP_LIST_FOLD_LEFT_691(o, s, BOOST_PP_LIST_REVERSE_D(691, l))
# define BOOST_PP_LIST_FOLD_RIGHT_692(o, s, l) BOOST_PP_LIST_FOLD_LEFT_692(o, s, BOOST_PP_LIST_REVERSE_D(692, l))
# define BOOST_PP_LIST_FOLD_RIGHT_693(o, s, l) BOOST_PP_LIST_FOLD_LEFT_693(o, s, BOOST_PP_LIST_REVERSE_D(693, l))
# define BOOST_PP_LIST_FOLD_RIGHT_694(o, s, l) BOOST_PP_LIST_FOLD_LEFT_694(o, s, BOOST_PP_LIST_REVERSE_D(694, l))
# define BOOST_PP_LIST_FOLD_RIGHT_695(o, s, l) BOOST_PP_LIST_FOLD_LEFT_695(o, s, BOOST_PP_LIST_REVERSE_D(695, l))
# define BOOST_PP_LIST_FOLD_RIGHT_696(o, s, l) BOOST_PP_LIST_FOLD_LEFT_696(o, s, BOOST_PP_LIST_REVERSE_D(696, l))
# define BOOST_PP_LIST_FOLD_RIGHT_697(o, s, l) BOOST_PP_LIST_FOLD_LEFT_697(o, s, BOOST_PP_LIST_REVERSE_D(697, l))
# define BOOST_PP_LIST_FOLD_RIGHT_698(o, s, l) BOOST_PP_LIST_FOLD_LEFT_698(o, s, BOOST_PP_LIST_REVERSE_D(698, l))
# define BOOST_PP_LIST_FOLD_RIGHT_699(o, s, l) BOOST_PP_LIST_FOLD_LEFT_699(o, s, BOOST_PP_LIST_REVERSE_D(699, l))
# define BOOST_PP_LIST_FOLD_RIGHT_700(o, s, l) BOOST_PP_LIST_FOLD_LEFT_700(o, s, BOOST_PP_LIST_REVERSE_D(700, l))
# define BOOST_PP_LIST_FOLD_RIGHT_701(o, s, l) BOOST_PP_LIST_FOLD_LEFT_701(o, s, BOOST_PP_LIST_REVERSE_D(701, l))
# define BOOST_PP_LIST_FOLD_RIGHT_702(o, s, l) BOOST_PP_LIST_FOLD_LEFT_702(o, s, BOOST_PP_LIST_REVERSE_D(702, l))
# define BOOST_PP_LIST_FOLD_RIGHT_703(o, s, l) BOOST_PP_LIST_FOLD_LEFT_703(o, s, BOOST_PP_LIST_REVERSE_D(703, l))
# define BOOST_PP_LIST_FOLD_RIGHT_704(o, s, l) BOOST_PP_LIST_FOLD_LEFT_704(o, s, BOOST_PP_LIST_REVERSE_D(704, l))
# define BOOST_PP_LIST_FOLD_RIGHT_705(o, s, l) BOOST_PP_LIST_FOLD_LEFT_705(o, s, BOOST_PP_LIST_REVERSE_D(705, l))
# define BOOST_PP_LIST_FOLD_RIGHT_706(o, s, l) BOOST_PP_LIST_FOLD_LEFT_706(o, s, BOOST_PP_LIST_REVERSE_D(706, l))
# define BOOST_PP_LIST_FOLD_RIGHT_707(o, s, l) BOOST_PP_LIST_FOLD_LEFT_707(o, s, BOOST_PP_LIST_REVERSE_D(707, l))
# define BOOST_PP_LIST_FOLD_RIGHT_708(o, s, l) BOOST_PP_LIST_FOLD_LEFT_708(o, s, BOOST_PP_LIST_REVERSE_D(708, l))
# define BOOST_PP_LIST_FOLD_RIGHT_709(o, s, l) BOOST_PP_LIST_FOLD_LEFT_709(o, s, BOOST_PP_LIST_REVERSE_D(709, l))
# define BOOST_PP_LIST_FOLD_RIGHT_710(o, s, l) BOOST_PP_LIST_FOLD_LEFT_710(o, s, BOOST_PP_LIST_REVERSE_D(710, l))
# define BOOST_PP_LIST_FOLD_RIGHT_711(o, s, l) BOOST_PP_LIST_FOLD_LEFT_711(o, s, BOOST_PP_LIST_REVERSE_D(711, l))
# define BOOST_PP_LIST_FOLD_RIGHT_712(o, s, l) BOOST_PP_LIST_FOLD_LEFT_712(o, s, BOOST_PP_LIST_REVERSE_D(712, l))
# define BOOST_PP_LIST_FOLD_RIGHT_713(o, s, l) BOOST_PP_LIST_FOLD_LEFT_713(o, s, BOOST_PP_LIST_REVERSE_D(713, l))
# define BOOST_PP_LIST_FOLD_RIGHT_714(o, s, l) BOOST_PP_LIST_FOLD_LEFT_714(o, s, BOOST_PP_LIST_REVERSE_D(714, l))
# define BOOST_PP_LIST_FOLD_RIGHT_715(o, s, l) BOOST_PP_LIST_FOLD_LEFT_715(o, s, BOOST_PP_LIST_REVERSE_D(715, l))
# define BOOST_PP_LIST_FOLD_RIGHT_716(o, s, l) BOOST_PP_LIST_FOLD_LEFT_716(o, s, BOOST_PP_LIST_REVERSE_D(716, l))
# define BOOST_PP_LIST_FOLD_RIGHT_717(o, s, l) BOOST_PP_LIST_FOLD_LEFT_717(o, s, BOOST_PP_LIST_REVERSE_D(717, l))
# define BOOST_PP_LIST_FOLD_RIGHT_718(o, s, l) BOOST_PP_LIST_FOLD_LEFT_718(o, s, BOOST_PP_LIST_REVERSE_D(718, l))
# define BOOST_PP_LIST_FOLD_RIGHT_719(o, s, l) BOOST_PP_LIST_FOLD_LEFT_719(o, s, BOOST_PP_LIST_REVERSE_D(719, l))
# define BOOST_PP_LIST_FOLD_RIGHT_720(o, s, l) BOOST_PP_LIST_FOLD_LEFT_720(o, s, BOOST_PP_LIST_REVERSE_D(720, l))
# define BOOST_PP_LIST_FOLD_RIGHT_721(o, s, l) BOOST_PP_LIST_FOLD_LEFT_721(o, s, BOOST_PP_LIST_REVERSE_D(721, l))
# define BOOST_PP_LIST_FOLD_RIGHT_722(o, s, l) BOOST_PP_LIST_FOLD_LEFT_722(o, s, BOOST_PP_LIST_REVERSE_D(722, l))
# define BOOST_PP_LIST_FOLD_RIGHT_723(o, s, l) BOOST_PP_LIST_FOLD_LEFT_723(o, s, BOOST_PP_LIST_REVERSE_D(723, l))
# define BOOST_PP_LIST_FOLD_RIGHT_724(o, s, l) BOOST_PP_LIST_FOLD_LEFT_724(o, s, BOOST_PP_LIST_REVERSE_D(724, l))
# define BOOST_PP_LIST_FOLD_RIGHT_725(o, s, l) BOOST_PP_LIST_FOLD_LEFT_725(o, s, BOOST_PP_LIST_REVERSE_D(725, l))
# define BOOST_PP_LIST_FOLD_RIGHT_726(o, s, l) BOOST_PP_LIST_FOLD_LEFT_726(o, s, BOOST_PP_LIST_REVERSE_D(726, l))
# define BOOST_PP_LIST_FOLD_RIGHT_727(o, s, l) BOOST_PP_LIST_FOLD_LEFT_727(o, s, BOOST_PP_LIST_REVERSE_D(727, l))
# define BOOST_PP_LIST_FOLD_RIGHT_728(o, s, l) BOOST_PP_LIST_FOLD_LEFT_728(o, s, BOOST_PP_LIST_REVERSE_D(728, l))
# define BOOST_PP_LIST_FOLD_RIGHT_729(o, s, l) BOOST_PP_LIST_FOLD_LEFT_729(o, s, BOOST_PP_LIST_REVERSE_D(729, l))
# define BOOST_PP_LIST_FOLD_RIGHT_730(o, s, l) BOOST_PP_LIST_FOLD_LEFT_730(o, s, BOOST_PP_LIST_REVERSE_D(730, l))
# define BOOST_PP_LIST_FOLD_RIGHT_731(o, s, l) BOOST_PP_LIST_FOLD_LEFT_731(o, s, BOOST_PP_LIST_REVERSE_D(731, l))
# define BOOST_PP_LIST_FOLD_RIGHT_732(o, s, l) BOOST_PP_LIST_FOLD_LEFT_732(o, s, BOOST_PP_LIST_REVERSE_D(732, l))
# define BOOST_PP_LIST_FOLD_RIGHT_733(o, s, l) BOOST_PP_LIST_FOLD_LEFT_733(o, s, BOOST_PP_LIST_REVERSE_D(733, l))
# define BOOST_PP_LIST_FOLD_RIGHT_734(o, s, l) BOOST_PP_LIST_FOLD_LEFT_734(o, s, BOOST_PP_LIST_REVERSE_D(734, l))
# define BOOST_PP_LIST_FOLD_RIGHT_735(o, s, l) BOOST_PP_LIST_FOLD_LEFT_735(o, s, BOOST_PP_LIST_REVERSE_D(735, l))
# define BOOST_PP_LIST_FOLD_RIGHT_736(o, s, l) BOOST_PP_LIST_FOLD_LEFT_736(o, s, BOOST_PP_LIST_REVERSE_D(736, l))
# define BOOST_PP_LIST_FOLD_RIGHT_737(o, s, l) BOOST_PP_LIST_FOLD_LEFT_737(o, s, BOOST_PP_LIST_REVERSE_D(737, l))
# define BOOST_PP_LIST_FOLD_RIGHT_738(o, s, l) BOOST_PP_LIST_FOLD_LEFT_738(o, s, BOOST_PP_LIST_REVERSE_D(738, l))
# define BOOST_PP_LIST_FOLD_RIGHT_739(o, s, l) BOOST_PP_LIST_FOLD_LEFT_739(o, s, BOOST_PP_LIST_REVERSE_D(739, l))
# define BOOST_PP_LIST_FOLD_RIGHT_740(o, s, l) BOOST_PP_LIST_FOLD_LEFT_740(o, s, BOOST_PP_LIST_REVERSE_D(740, l))
# define BOOST_PP_LIST_FOLD_RIGHT_741(o, s, l) BOOST_PP_LIST_FOLD_LEFT_741(o, s, BOOST_PP_LIST_REVERSE_D(741, l))
# define BOOST_PP_LIST_FOLD_RIGHT_742(o, s, l) BOOST_PP_LIST_FOLD_LEFT_742(o, s, BOOST_PP_LIST_REVERSE_D(742, l))
# define BOOST_PP_LIST_FOLD_RIGHT_743(o, s, l) BOOST_PP_LIST_FOLD_LEFT_743(o, s, BOOST_PP_LIST_REVERSE_D(743, l))
# define BOOST_PP_LIST_FOLD_RIGHT_744(o, s, l) BOOST_PP_LIST_FOLD_LEFT_744(o, s, BOOST_PP_LIST_REVERSE_D(744, l))
# define BOOST_PP_LIST_FOLD_RIGHT_745(o, s, l) BOOST_PP_LIST_FOLD_LEFT_745(o, s, BOOST_PP_LIST_REVERSE_D(745, l))
# define BOOST_PP_LIST_FOLD_RIGHT_746(o, s, l) BOOST_PP_LIST_FOLD_LEFT_746(o, s, BOOST_PP_LIST_REVERSE_D(746, l))
# define BOOST_PP_LIST_FOLD_RIGHT_747(o, s, l) BOOST_PP_LIST_FOLD_LEFT_747(o, s, BOOST_PP_LIST_REVERSE_D(747, l))
# define BOOST_PP_LIST_FOLD_RIGHT_748(o, s, l) BOOST_PP_LIST_FOLD_LEFT_748(o, s, BOOST_PP_LIST_REVERSE_D(748, l))
# define BOOST_PP_LIST_FOLD_RIGHT_749(o, s, l) BOOST_PP_LIST_FOLD_LEFT_749(o, s, BOOST_PP_LIST_REVERSE_D(749, l))
# define BOOST_PP_LIST_FOLD_RIGHT_750(o, s, l) BOOST_PP_LIST_FOLD_LEFT_750(o, s, BOOST_PP_LIST_REVERSE_D(750, l))
# define BOOST_PP_LIST_FOLD_RIGHT_751(o, s, l) BOOST_PP_LIST_FOLD_LEFT_751(o, s, BOOST_PP_LIST_REVERSE_D(751, l))
# define BOOST_PP_LIST_FOLD_RIGHT_752(o, s, l) BOOST_PP_LIST_FOLD_LEFT_752(o, s, BOOST_PP_LIST_REVERSE_D(752, l))
# define BOOST_PP_LIST_FOLD_RIGHT_753(o, s, l) BOOST_PP_LIST_FOLD_LEFT_753(o, s, BOOST_PP_LIST_REVERSE_D(753, l))
# define BOOST_PP_LIST_FOLD_RIGHT_754(o, s, l) BOOST_PP_LIST_FOLD_LEFT_754(o, s, BOOST_PP_LIST_REVERSE_D(754, l))
# define BOOST_PP_LIST_FOLD_RIGHT_755(o, s, l) BOOST_PP_LIST_FOLD_LEFT_755(o, s, BOOST_PP_LIST_REVERSE_D(755, l))
# define BOOST_PP_LIST_FOLD_RIGHT_756(o, s, l) BOOST_PP_LIST_FOLD_LEFT_756(o, s, BOOST_PP_LIST_REVERSE_D(756, l))
# define BOOST_PP_LIST_FOLD_RIGHT_757(o, s, l) BOOST_PP_LIST_FOLD_LEFT_757(o, s, BOOST_PP_LIST_REVERSE_D(757, l))
# define BOOST_PP_LIST_FOLD_RIGHT_758(o, s, l) BOOST_PP_LIST_FOLD_LEFT_758(o, s, BOOST_PP_LIST_REVERSE_D(758, l))
# define BOOST_PP_LIST_FOLD_RIGHT_759(o, s, l) BOOST_PP_LIST_FOLD_LEFT_759(o, s, BOOST_PP_LIST_REVERSE_D(759, l))
# define BOOST_PP_LIST_FOLD_RIGHT_760(o, s, l) BOOST_PP_LIST_FOLD_LEFT_760(o, s, BOOST_PP_LIST_REVERSE_D(760, l))
# define BOOST_PP_LIST_FOLD_RIGHT_761(o, s, l) BOOST_PP_LIST_FOLD_LEFT_761(o, s, BOOST_PP_LIST_REVERSE_D(761, l))
# define BOOST_PP_LIST_FOLD_RIGHT_762(o, s, l) BOOST_PP_LIST_FOLD_LEFT_762(o, s, BOOST_PP_LIST_REVERSE_D(762, l))
# define BOOST_PP_LIST_FOLD_RIGHT_763(o, s, l) BOOST_PP_LIST_FOLD_LEFT_763(o, s, BOOST_PP_LIST_REVERSE_D(763, l))
# define BOOST_PP_LIST_FOLD_RIGHT_764(o, s, l) BOOST_PP_LIST_FOLD_LEFT_764(o, s, BOOST_PP_LIST_REVERSE_D(764, l))
# define BOOST_PP_LIST_FOLD_RIGHT_765(o, s, l) BOOST_PP_LIST_FOLD_LEFT_765(o, s, BOOST_PP_LIST_REVERSE_D(765, l))
# define BOOST_PP_LIST_FOLD_RIGHT_766(o, s, l) BOOST_PP_LIST_FOLD_LEFT_766(o, s, BOOST_PP_LIST_REVERSE_D(766, l))
# define BOOST_PP_LIST_FOLD_RIGHT_767(o, s, l) BOOST_PP_LIST_FOLD_LEFT_767(o, s, BOOST_PP_LIST_REVERSE_D(767, l))
# define BOOST_PP_LIST_FOLD_RIGHT_768(o, s, l) BOOST_PP_LIST_FOLD_LEFT_768(o, s, BOOST_PP_LIST_REVERSE_D(768, l))
# define BOOST_PP_LIST_FOLD_RIGHT_769(o, s, l) BOOST_PP_LIST_FOLD_LEFT_769(o, s, BOOST_PP_LIST_REVERSE_D(769, l))
# define BOOST_PP_LIST_FOLD_RIGHT_770(o, s, l) BOOST_PP_LIST_FOLD_LEFT_770(o, s, BOOST_PP_LIST_REVERSE_D(770, l))
# define BOOST_PP_LIST_FOLD_RIGHT_771(o, s, l) BOOST_PP_LIST_FOLD_LEFT_771(o, s, BOOST_PP_LIST_REVERSE_D(771, l))
# define BOOST_PP_LIST_FOLD_RIGHT_772(o, s, l) BOOST_PP_LIST_FOLD_LEFT_772(o, s, BOOST_PP_LIST_REVERSE_D(772, l))
# define BOOST_PP_LIST_FOLD_RIGHT_773(o, s, l) BOOST_PP_LIST_FOLD_LEFT_773(o, s, BOOST_PP_LIST_REVERSE_D(773, l))
# define BOOST_PP_LIST_FOLD_RIGHT_774(o, s, l) BOOST_PP_LIST_FOLD_LEFT_774(o, s, BOOST_PP_LIST_REVERSE_D(774, l))
# define BOOST_PP_LIST_FOLD_RIGHT_775(o, s, l) BOOST_PP_LIST_FOLD_LEFT_775(o, s, BOOST_PP_LIST_REVERSE_D(775, l))
# define BOOST_PP_LIST_FOLD_RIGHT_776(o, s, l) BOOST_PP_LIST_FOLD_LEFT_776(o, s, BOOST_PP_LIST_REVERSE_D(776, l))
# define BOOST_PP_LIST_FOLD_RIGHT_777(o, s, l) BOOST_PP_LIST_FOLD_LEFT_777(o, s, BOOST_PP_LIST_REVERSE_D(777, l))
# define BOOST_PP_LIST_FOLD_RIGHT_778(o, s, l) BOOST_PP_LIST_FOLD_LEFT_778(o, s, BOOST_PP_LIST_REVERSE_D(778, l))
# define BOOST_PP_LIST_FOLD_RIGHT_779(o, s, l) BOOST_PP_LIST_FOLD_LEFT_779(o, s, BOOST_PP_LIST_REVERSE_D(779, l))
# define BOOST_PP_LIST_FOLD_RIGHT_780(o, s, l) BOOST_PP_LIST_FOLD_LEFT_780(o, s, BOOST_PP_LIST_REVERSE_D(780, l))
# define BOOST_PP_LIST_FOLD_RIGHT_781(o, s, l) BOOST_PP_LIST_FOLD_LEFT_781(o, s, BOOST_PP_LIST_REVERSE_D(781, l))
# define BOOST_PP_LIST_FOLD_RIGHT_782(o, s, l) BOOST_PP_LIST_FOLD_LEFT_782(o, s, BOOST_PP_LIST_REVERSE_D(782, l))
# define BOOST_PP_LIST_FOLD_RIGHT_783(o, s, l) BOOST_PP_LIST_FOLD_LEFT_783(o, s, BOOST_PP_LIST_REVERSE_D(783, l))
# define BOOST_PP_LIST_FOLD_RIGHT_784(o, s, l) BOOST_PP_LIST_FOLD_LEFT_784(o, s, BOOST_PP_LIST_REVERSE_D(784, l))
# define BOOST_PP_LIST_FOLD_RIGHT_785(o, s, l) BOOST_PP_LIST_FOLD_LEFT_785(o, s, BOOST_PP_LIST_REVERSE_D(785, l))
# define BOOST_PP_LIST_FOLD_RIGHT_786(o, s, l) BOOST_PP_LIST_FOLD_LEFT_786(o, s, BOOST_PP_LIST_REVERSE_D(786, l))
# define BOOST_PP_LIST_FOLD_RIGHT_787(o, s, l) BOOST_PP_LIST_FOLD_LEFT_787(o, s, BOOST_PP_LIST_REVERSE_D(787, l))
# define BOOST_PP_LIST_FOLD_RIGHT_788(o, s, l) BOOST_PP_LIST_FOLD_LEFT_788(o, s, BOOST_PP_LIST_REVERSE_D(788, l))
# define BOOST_PP_LIST_FOLD_RIGHT_789(o, s, l) BOOST_PP_LIST_FOLD_LEFT_789(o, s, BOOST_PP_LIST_REVERSE_D(789, l))
# define BOOST_PP_LIST_FOLD_RIGHT_790(o, s, l) BOOST_PP_LIST_FOLD_LEFT_790(o, s, BOOST_PP_LIST_REVERSE_D(790, l))
# define BOOST_PP_LIST_FOLD_RIGHT_791(o, s, l) BOOST_PP_LIST_FOLD_LEFT_791(o, s, BOOST_PP_LIST_REVERSE_D(791, l))
# define BOOST_PP_LIST_FOLD_RIGHT_792(o, s, l) BOOST_PP_LIST_FOLD_LEFT_792(o, s, BOOST_PP_LIST_REVERSE_D(792, l))
# define BOOST_PP_LIST_FOLD_RIGHT_793(o, s, l) BOOST_PP_LIST_FOLD_LEFT_793(o, s, BOOST_PP_LIST_REVERSE_D(793, l))
# define BOOST_PP_LIST_FOLD_RIGHT_794(o, s, l) BOOST_PP_LIST_FOLD_LEFT_794(o, s, BOOST_PP_LIST_REVERSE_D(794, l))
# define BOOST_PP_LIST_FOLD_RIGHT_795(o, s, l) BOOST_PP_LIST_FOLD_LEFT_795(o, s, BOOST_PP_LIST_REVERSE_D(795, l))
# define BOOST_PP_LIST_FOLD_RIGHT_796(o, s, l) BOOST_PP_LIST_FOLD_LEFT_796(o, s, BOOST_PP_LIST_REVERSE_D(796, l))
# define BOOST_PP_LIST_FOLD_RIGHT_797(o, s, l) BOOST_PP_LIST_FOLD_LEFT_797(o, s, BOOST_PP_LIST_REVERSE_D(797, l))
# define BOOST_PP_LIST_FOLD_RIGHT_798(o, s, l) BOOST_PP_LIST_FOLD_LEFT_798(o, s, BOOST_PP_LIST_REVERSE_D(798, l))
# define BOOST_PP_LIST_FOLD_RIGHT_799(o, s, l) BOOST_PP_LIST_FOLD_LEFT_799(o, s, BOOST_PP_LIST_REVERSE_D(799, l))
# define BOOST_PP_LIST_FOLD_RIGHT_800(o, s, l) BOOST_PP_LIST_FOLD_LEFT_800(o, s, BOOST_PP_LIST_REVERSE_D(800, l))
# define BOOST_PP_LIST_FOLD_RIGHT_801(o, s, l) BOOST_PP_LIST_FOLD_LEFT_801(o, s, BOOST_PP_LIST_REVERSE_D(801, l))
# define BOOST_PP_LIST_FOLD_RIGHT_802(o, s, l) BOOST_PP_LIST_FOLD_LEFT_802(o, s, BOOST_PP_LIST_REVERSE_D(802, l))
# define BOOST_PP_LIST_FOLD_RIGHT_803(o, s, l) BOOST_PP_LIST_FOLD_LEFT_803(o, s, BOOST_PP_LIST_REVERSE_D(803, l))
# define BOOST_PP_LIST_FOLD_RIGHT_804(o, s, l) BOOST_PP_LIST_FOLD_LEFT_804(o, s, BOOST_PP_LIST_REVERSE_D(804, l))
# define BOOST_PP_LIST_FOLD_RIGHT_805(o, s, l) BOOST_PP_LIST_FOLD_LEFT_805(o, s, BOOST_PP_LIST_REVERSE_D(805, l))
# define BOOST_PP_LIST_FOLD_RIGHT_806(o, s, l) BOOST_PP_LIST_FOLD_LEFT_806(o, s, BOOST_PP_LIST_REVERSE_D(806, l))
# define BOOST_PP_LIST_FOLD_RIGHT_807(o, s, l) BOOST_PP_LIST_FOLD_LEFT_807(o, s, BOOST_PP_LIST_REVERSE_D(807, l))
# define BOOST_PP_LIST_FOLD_RIGHT_808(o, s, l) BOOST_PP_LIST_FOLD_LEFT_808(o, s, BOOST_PP_LIST_REVERSE_D(808, l))
# define BOOST_PP_LIST_FOLD_RIGHT_809(o, s, l) BOOST_PP_LIST_FOLD_LEFT_809(o, s, BOOST_PP_LIST_REVERSE_D(809, l))
# define BOOST_PP_LIST_FOLD_RIGHT_810(o, s, l) BOOST_PP_LIST_FOLD_LEFT_810(o, s, BOOST_PP_LIST_REVERSE_D(810, l))
# define BOOST_PP_LIST_FOLD_RIGHT_811(o, s, l) BOOST_PP_LIST_FOLD_LEFT_811(o, s, BOOST_PP_LIST_REVERSE_D(811, l))
# define BOOST_PP_LIST_FOLD_RIGHT_812(o, s, l) BOOST_PP_LIST_FOLD_LEFT_812(o, s, BOOST_PP_LIST_REVERSE_D(812, l))
# define BOOST_PP_LIST_FOLD_RIGHT_813(o, s, l) BOOST_PP_LIST_FOLD_LEFT_813(o, s, BOOST_PP_LIST_REVERSE_D(813, l))
# define BOOST_PP_LIST_FOLD_RIGHT_814(o, s, l) BOOST_PP_LIST_FOLD_LEFT_814(o, s, BOOST_PP_LIST_REVERSE_D(814, l))
# define BOOST_PP_LIST_FOLD_RIGHT_815(o, s, l) BOOST_PP_LIST_FOLD_LEFT_815(o, s, BOOST_PP_LIST_REVERSE_D(815, l))
# define BOOST_PP_LIST_FOLD_RIGHT_816(o, s, l) BOOST_PP_LIST_FOLD_LEFT_816(o, s, BOOST_PP_LIST_REVERSE_D(816, l))
# define BOOST_PP_LIST_FOLD_RIGHT_817(o, s, l) BOOST_PP_LIST_FOLD_LEFT_817(o, s, BOOST_PP_LIST_REVERSE_D(817, l))
# define BOOST_PP_LIST_FOLD_RIGHT_818(o, s, l) BOOST_PP_LIST_FOLD_LEFT_818(o, s, BOOST_PP_LIST_REVERSE_D(818, l))
# define BOOST_PP_LIST_FOLD_RIGHT_819(o, s, l) BOOST_PP_LIST_FOLD_LEFT_819(o, s, BOOST_PP_LIST_REVERSE_D(819, l))
# define BOOST_PP_LIST_FOLD_RIGHT_820(o, s, l) BOOST_PP_LIST_FOLD_LEFT_820(o, s, BOOST_PP_LIST_REVERSE_D(820, l))
# define BOOST_PP_LIST_FOLD_RIGHT_821(o, s, l) BOOST_PP_LIST_FOLD_LEFT_821(o, s, BOOST_PP_LIST_REVERSE_D(821, l))
# define BOOST_PP_LIST_FOLD_RIGHT_822(o, s, l) BOOST_PP_LIST_FOLD_LEFT_822(o, s, BOOST_PP_LIST_REVERSE_D(822, l))
# define BOOST_PP_LIST_FOLD_RIGHT_823(o, s, l) BOOST_PP_LIST_FOLD_LEFT_823(o, s, BOOST_PP_LIST_REVERSE_D(823, l))
# define BOOST_PP_LIST_FOLD_RIGHT_824(o, s, l) BOOST_PP_LIST_FOLD_LEFT_824(o, s, BOOST_PP_LIST_REVERSE_D(824, l))
# define BOOST_PP_LIST_FOLD_RIGHT_825(o, s, l) BOOST_PP_LIST_FOLD_LEFT_825(o, s, BOOST_PP_LIST_REVERSE_D(825, l))
# define BOOST_PP_LIST_FOLD_RIGHT_826(o, s, l) BOOST_PP_LIST_FOLD_LEFT_826(o, s, BOOST_PP_LIST_REVERSE_D(826, l))
# define BOOST_PP_LIST_FOLD_RIGHT_827(o, s, l) BOOST_PP_LIST_FOLD_LEFT_827(o, s, BOOST_PP_LIST_REVERSE_D(827, l))
# define BOOST_PP_LIST_FOLD_RIGHT_828(o, s, l) BOOST_PP_LIST_FOLD_LEFT_828(o, s, BOOST_PP_LIST_REVERSE_D(828, l))
# define BOOST_PP_LIST_FOLD_RIGHT_829(o, s, l) BOOST_PP_LIST_FOLD_LEFT_829(o, s, BOOST_PP_LIST_REVERSE_D(829, l))
# define BOOST_PP_LIST_FOLD_RIGHT_830(o, s, l) BOOST_PP_LIST_FOLD_LEFT_830(o, s, BOOST_PP_LIST_REVERSE_D(830, l))
# define BOOST_PP_LIST_FOLD_RIGHT_831(o, s, l) BOOST_PP_LIST_FOLD_LEFT_831(o, s, BOOST_PP_LIST_REVERSE_D(831, l))
# define BOOST_PP_LIST_FOLD_RIGHT_832(o, s, l) BOOST_PP_LIST_FOLD_LEFT_832(o, s, BOOST_PP_LIST_REVERSE_D(832, l))
# define BOOST_PP_LIST_FOLD_RIGHT_833(o, s, l) BOOST_PP_LIST_FOLD_LEFT_833(o, s, BOOST_PP_LIST_REVERSE_D(833, l))
# define BOOST_PP_LIST_FOLD_RIGHT_834(o, s, l) BOOST_PP_LIST_FOLD_LEFT_834(o, s, BOOST_PP_LIST_REVERSE_D(834, l))
# define BOOST_PP_LIST_FOLD_RIGHT_835(o, s, l) BOOST_PP_LIST_FOLD_LEFT_835(o, s, BOOST_PP_LIST_REVERSE_D(835, l))
# define BOOST_PP_LIST_FOLD_RIGHT_836(o, s, l) BOOST_PP_LIST_FOLD_LEFT_836(o, s, BOOST_PP_LIST_REVERSE_D(836, l))
# define BOOST_PP_LIST_FOLD_RIGHT_837(o, s, l) BOOST_PP_LIST_FOLD_LEFT_837(o, s, BOOST_PP_LIST_REVERSE_D(837, l))
# define BOOST_PP_LIST_FOLD_RIGHT_838(o, s, l) BOOST_PP_LIST_FOLD_LEFT_838(o, s, BOOST_PP_LIST_REVERSE_D(838, l))
# define BOOST_PP_LIST_FOLD_RIGHT_839(o, s, l) BOOST_PP_LIST_FOLD_LEFT_839(o, s, BOOST_PP_LIST_REVERSE_D(839, l))
# define BOOST_PP_LIST_FOLD_RIGHT_840(o, s, l) BOOST_PP_LIST_FOLD_LEFT_840(o, s, BOOST_PP_LIST_REVERSE_D(840, l))
# define BOOST_PP_LIST_FOLD_RIGHT_841(o, s, l) BOOST_PP_LIST_FOLD_LEFT_841(o, s, BOOST_PP_LIST_REVERSE_D(841, l))
# define BOOST_PP_LIST_FOLD_RIGHT_842(o, s, l) BOOST_PP_LIST_FOLD_LEFT_842(o, s, BOOST_PP_LIST_REVERSE_D(842, l))
# define BOOST_PP_LIST_FOLD_RIGHT_843(o, s, l) BOOST_PP_LIST_FOLD_LEFT_843(o, s, BOOST_PP_LIST_REVERSE_D(843, l))
# define BOOST_PP_LIST_FOLD_RIGHT_844(o, s, l) BOOST_PP_LIST_FOLD_LEFT_844(o, s, BOOST_PP_LIST_REVERSE_D(844, l))
# define BOOST_PP_LIST_FOLD_RIGHT_845(o, s, l) BOOST_PP_LIST_FOLD_LEFT_845(o, s, BOOST_PP_LIST_REVERSE_D(845, l))
# define BOOST_PP_LIST_FOLD_RIGHT_846(o, s, l) BOOST_PP_LIST_FOLD_LEFT_846(o, s, BOOST_PP_LIST_REVERSE_D(846, l))
# define BOOST_PP_LIST_FOLD_RIGHT_847(o, s, l) BOOST_PP_LIST_FOLD_LEFT_847(o, s, BOOST_PP_LIST_REVERSE_D(847, l))
# define BOOST_PP_LIST_FOLD_RIGHT_848(o, s, l) BOOST_PP_LIST_FOLD_LEFT_848(o, s, BOOST_PP_LIST_REVERSE_D(848, l))
# define BOOST_PP_LIST_FOLD_RIGHT_849(o, s, l) BOOST_PP_LIST_FOLD_LEFT_849(o, s, BOOST_PP_LIST_REVERSE_D(849, l))
# define BOOST_PP_LIST_FOLD_RIGHT_850(o, s, l) BOOST_PP_LIST_FOLD_LEFT_850(o, s, BOOST_PP_LIST_REVERSE_D(850, l))
# define BOOST_PP_LIST_FOLD_RIGHT_851(o, s, l) BOOST_PP_LIST_FOLD_LEFT_851(o, s, BOOST_PP_LIST_REVERSE_D(851, l))
# define BOOST_PP_LIST_FOLD_RIGHT_852(o, s, l) BOOST_PP_LIST_FOLD_LEFT_852(o, s, BOOST_PP_LIST_REVERSE_D(852, l))
# define BOOST_PP_LIST_FOLD_RIGHT_853(o, s, l) BOOST_PP_LIST_FOLD_LEFT_853(o, s, BOOST_PP_LIST_REVERSE_D(853, l))
# define BOOST_PP_LIST_FOLD_RIGHT_854(o, s, l) BOOST_PP_LIST_FOLD_LEFT_854(o, s, BOOST_PP_LIST_REVERSE_D(854, l))
# define BOOST_PP_LIST_FOLD_RIGHT_855(o, s, l) BOOST_PP_LIST_FOLD_LEFT_855(o, s, BOOST_PP_LIST_REVERSE_D(855, l))
# define BOOST_PP_LIST_FOLD_RIGHT_856(o, s, l) BOOST_PP_LIST_FOLD_LEFT_856(o, s, BOOST_PP_LIST_REVERSE_D(856, l))
# define BOOST_PP_LIST_FOLD_RIGHT_857(o, s, l) BOOST_PP_LIST_FOLD_LEFT_857(o, s, BOOST_PP_LIST_REVERSE_D(857, l))
# define BOOST_PP_LIST_FOLD_RIGHT_858(o, s, l) BOOST_PP_LIST_FOLD_LEFT_858(o, s, BOOST_PP_LIST_REVERSE_D(858, l))
# define BOOST_PP_LIST_FOLD_RIGHT_859(o, s, l) BOOST_PP_LIST_FOLD_LEFT_859(o, s, BOOST_PP_LIST_REVERSE_D(859, l))
# define BOOST_PP_LIST_FOLD_RIGHT_860(o, s, l) BOOST_PP_LIST_FOLD_LEFT_860(o, s, BOOST_PP_LIST_REVERSE_D(860, l))
# define BOOST_PP_LIST_FOLD_RIGHT_861(o, s, l) BOOST_PP_LIST_FOLD_LEFT_861(o, s, BOOST_PP_LIST_REVERSE_D(861, l))
# define BOOST_PP_LIST_FOLD_RIGHT_862(o, s, l) BOOST_PP_LIST_FOLD_LEFT_862(o, s, BOOST_PP_LIST_REVERSE_D(862, l))
# define BOOST_PP_LIST_FOLD_RIGHT_863(o, s, l) BOOST_PP_LIST_FOLD_LEFT_863(o, s, BOOST_PP_LIST_REVERSE_D(863, l))
# define BOOST_PP_LIST_FOLD_RIGHT_864(o, s, l) BOOST_PP_LIST_FOLD_LEFT_864(o, s, BOOST_PP_LIST_REVERSE_D(864, l))
# define BOOST_PP_LIST_FOLD_RIGHT_865(o, s, l) BOOST_PP_LIST_FOLD_LEFT_865(o, s, BOOST_PP_LIST_REVERSE_D(865, l))
# define BOOST_PP_LIST_FOLD_RIGHT_866(o, s, l) BOOST_PP_LIST_FOLD_LEFT_866(o, s, BOOST_PP_LIST_REVERSE_D(866, l))
# define BOOST_PP_LIST_FOLD_RIGHT_867(o, s, l) BOOST_PP_LIST_FOLD_LEFT_867(o, s, BOOST_PP_LIST_REVERSE_D(867, l))
# define BOOST_PP_LIST_FOLD_RIGHT_868(o, s, l) BOOST_PP_LIST_FOLD_LEFT_868(o, s, BOOST_PP_LIST_REVERSE_D(868, l))
# define BOOST_PP_LIST_FOLD_RIGHT_869(o, s, l) BOOST_PP_LIST_FOLD_LEFT_869(o, s, BOOST_PP_LIST_REVERSE_D(869, l))
# define BOOST_PP_LIST_FOLD_RIGHT_870(o, s, l) BOOST_PP_LIST_FOLD_LEFT_870(o, s, BOOST_PP_LIST_REVERSE_D(870, l))
# define BOOST_PP_LIST_FOLD_RIGHT_871(o, s, l) BOOST_PP_LIST_FOLD_LEFT_871(o, s, BOOST_PP_LIST_REVERSE_D(871, l))
# define BOOST_PP_LIST_FOLD_RIGHT_872(o, s, l) BOOST_PP_LIST_FOLD_LEFT_872(o, s, BOOST_PP_LIST_REVERSE_D(872, l))
# define BOOST_PP_LIST_FOLD_RIGHT_873(o, s, l) BOOST_PP_LIST_FOLD_LEFT_873(o, s, BOOST_PP_LIST_REVERSE_D(873, l))
# define BOOST_PP_LIST_FOLD_RIGHT_874(o, s, l) BOOST_PP_LIST_FOLD_LEFT_874(o, s, BOOST_PP_LIST_REVERSE_D(874, l))
# define BOOST_PP_LIST_FOLD_RIGHT_875(o, s, l) BOOST_PP_LIST_FOLD_LEFT_875(o, s, BOOST_PP_LIST_REVERSE_D(875, l))
# define BOOST_PP_LIST_FOLD_RIGHT_876(o, s, l) BOOST_PP_LIST_FOLD_LEFT_876(o, s, BOOST_PP_LIST_REVERSE_D(876, l))
# define BOOST_PP_LIST_FOLD_RIGHT_877(o, s, l) BOOST_PP_LIST_FOLD_LEFT_877(o, s, BOOST_PP_LIST_REVERSE_D(877, l))
# define BOOST_PP_LIST_FOLD_RIGHT_878(o, s, l) BOOST_PP_LIST_FOLD_LEFT_878(o, s, BOOST_PP_LIST_REVERSE_D(878, l))
# define BOOST_PP_LIST_FOLD_RIGHT_879(o, s, l) BOOST_PP_LIST_FOLD_LEFT_879(o, s, BOOST_PP_LIST_REVERSE_D(879, l))
# define BOOST_PP_LIST_FOLD_RIGHT_880(o, s, l) BOOST_PP_LIST_FOLD_LEFT_880(o, s, BOOST_PP_LIST_REVERSE_D(880, l))
# define BOOST_PP_LIST_FOLD_RIGHT_881(o, s, l) BOOST_PP_LIST_FOLD_LEFT_881(o, s, BOOST_PP_LIST_REVERSE_D(881, l))
# define BOOST_PP_LIST_FOLD_RIGHT_882(o, s, l) BOOST_PP_LIST_FOLD_LEFT_882(o, s, BOOST_PP_LIST_REVERSE_D(882, l))
# define BOOST_PP_LIST_FOLD_RIGHT_883(o, s, l) BOOST_PP_LIST_FOLD_LEFT_883(o, s, BOOST_PP_LIST_REVERSE_D(883, l))
# define BOOST_PP_LIST_FOLD_RIGHT_884(o, s, l) BOOST_PP_LIST_FOLD_LEFT_884(o, s, BOOST_PP_LIST_REVERSE_D(884, l))
# define BOOST_PP_LIST_FOLD_RIGHT_885(o, s, l) BOOST_PP_LIST_FOLD_LEFT_885(o, s, BOOST_PP_LIST_REVERSE_D(885, l))
# define BOOST_PP_LIST_FOLD_RIGHT_886(o, s, l) BOOST_PP_LIST_FOLD_LEFT_886(o, s, BOOST_PP_LIST_REVERSE_D(886, l))
# define BOOST_PP_LIST_FOLD_RIGHT_887(o, s, l) BOOST_PP_LIST_FOLD_LEFT_887(o, s, BOOST_PP_LIST_REVERSE_D(887, l))
# define BOOST_PP_LIST_FOLD_RIGHT_888(o, s, l) BOOST_PP_LIST_FOLD_LEFT_888(o, s, BOOST_PP_LIST_REVERSE_D(888, l))
# define BOOST_PP_LIST_FOLD_RIGHT_889(o, s, l) BOOST_PP_LIST_FOLD_LEFT_889(o, s, BOOST_PP_LIST_REVERSE_D(889, l))
# define BOOST_PP_LIST_FOLD_RIGHT_890(o, s, l) BOOST_PP_LIST_FOLD_LEFT_890(o, s, BOOST_PP_LIST_REVERSE_D(890, l))
# define BOOST_PP_LIST_FOLD_RIGHT_891(o, s, l) BOOST_PP_LIST_FOLD_LEFT_891(o, s, BOOST_PP_LIST_REVERSE_D(891, l))
# define BOOST_PP_LIST_FOLD_RIGHT_892(o, s, l) BOOST_PP_LIST_FOLD_LEFT_892(o, s, BOOST_PP_LIST_REVERSE_D(892, l))
# define BOOST_PP_LIST_FOLD_RIGHT_893(o, s, l) BOOST_PP_LIST_FOLD_LEFT_893(o, s, BOOST_PP_LIST_REVERSE_D(893, l))
# define BOOST_PP_LIST_FOLD_RIGHT_894(o, s, l) BOOST_PP_LIST_FOLD_LEFT_894(o, s, BOOST_PP_LIST_REVERSE_D(894, l))
# define BOOST_PP_LIST_FOLD_RIGHT_895(o, s, l) BOOST_PP_LIST_FOLD_LEFT_895(o, s, BOOST_PP_LIST_REVERSE_D(895, l))
# define BOOST_PP_LIST_FOLD_RIGHT_896(o, s, l) BOOST_PP_LIST_FOLD_LEFT_896(o, s, BOOST_PP_LIST_REVERSE_D(896, l))
# define BOOST_PP_LIST_FOLD_RIGHT_897(o, s, l) BOOST_PP_LIST_FOLD_LEFT_897(o, s, BOOST_PP_LIST_REVERSE_D(897, l))
# define BOOST_PP_LIST_FOLD_RIGHT_898(o, s, l) BOOST_PP_LIST_FOLD_LEFT_898(o, s, BOOST_PP_LIST_REVERSE_D(898, l))
# define BOOST_PP_LIST_FOLD_RIGHT_899(o, s, l) BOOST_PP_LIST_FOLD_LEFT_899(o, s, BOOST_PP_LIST_REVERSE_D(899, l))
# define BOOST_PP_LIST_FOLD_RIGHT_900(o, s, l) BOOST_PP_LIST_FOLD_LEFT_900(o, s, BOOST_PP_LIST_REVERSE_D(900, l))
# define BOOST_PP_LIST_FOLD_RIGHT_901(o, s, l) BOOST_PP_LIST_FOLD_LEFT_901(o, s, BOOST_PP_LIST_REVERSE_D(901, l))
# define BOOST_PP_LIST_FOLD_RIGHT_902(o, s, l) BOOST_PP_LIST_FOLD_LEFT_902(o, s, BOOST_PP_LIST_REVERSE_D(902, l))
# define BOOST_PP_LIST_FOLD_RIGHT_903(o, s, l) BOOST_PP_LIST_FOLD_LEFT_903(o, s, BOOST_PP_LIST_REVERSE_D(903, l))
# define BOOST_PP_LIST_FOLD_RIGHT_904(o, s, l) BOOST_PP_LIST_FOLD_LEFT_904(o, s, BOOST_PP_LIST_REVERSE_D(904, l))
# define BOOST_PP_LIST_FOLD_RIGHT_905(o, s, l) BOOST_PP_LIST_FOLD_LEFT_905(o, s, BOOST_PP_LIST_REVERSE_D(905, l))
# define BOOST_PP_LIST_FOLD_RIGHT_906(o, s, l) BOOST_PP_LIST_FOLD_LEFT_906(o, s, BOOST_PP_LIST_REVERSE_D(906, l))
# define BOOST_PP_LIST_FOLD_RIGHT_907(o, s, l) BOOST_PP_LIST_FOLD_LEFT_907(o, s, BOOST_PP_LIST_REVERSE_D(907, l))
# define BOOST_PP_LIST_FOLD_RIGHT_908(o, s, l) BOOST_PP_LIST_FOLD_LEFT_908(o, s, BOOST_PP_LIST_REVERSE_D(908, l))
# define BOOST_PP_LIST_FOLD_RIGHT_909(o, s, l) BOOST_PP_LIST_FOLD_LEFT_909(o, s, BOOST_PP_LIST_REVERSE_D(909, l))
# define BOOST_PP_LIST_FOLD_RIGHT_910(o, s, l) BOOST_PP_LIST_FOLD_LEFT_910(o, s, BOOST_PP_LIST_REVERSE_D(910, l))
# define BOOST_PP_LIST_FOLD_RIGHT_911(o, s, l) BOOST_PP_LIST_FOLD_LEFT_911(o, s, BOOST_PP_LIST_REVERSE_D(911, l))
# define BOOST_PP_LIST_FOLD_RIGHT_912(o, s, l) BOOST_PP_LIST_FOLD_LEFT_912(o, s, BOOST_PP_LIST_REVERSE_D(912, l))
# define BOOST_PP_LIST_FOLD_RIGHT_913(o, s, l) BOOST_PP_LIST_FOLD_LEFT_913(o, s, BOOST_PP_LIST_REVERSE_D(913, l))
# define BOOST_PP_LIST_FOLD_RIGHT_914(o, s, l) BOOST_PP_LIST_FOLD_LEFT_914(o, s, BOOST_PP_LIST_REVERSE_D(914, l))
# define BOOST_PP_LIST_FOLD_RIGHT_915(o, s, l) BOOST_PP_LIST_FOLD_LEFT_915(o, s, BOOST_PP_LIST_REVERSE_D(915, l))
# define BOOST_PP_LIST_FOLD_RIGHT_916(o, s, l) BOOST_PP_LIST_FOLD_LEFT_916(o, s, BOOST_PP_LIST_REVERSE_D(916, l))
# define BOOST_PP_LIST_FOLD_RIGHT_917(o, s, l) BOOST_PP_LIST_FOLD_LEFT_917(o, s, BOOST_PP_LIST_REVERSE_D(917, l))
# define BOOST_PP_LIST_FOLD_RIGHT_918(o, s, l) BOOST_PP_LIST_FOLD_LEFT_918(o, s, BOOST_PP_LIST_REVERSE_D(918, l))
# define BOOST_PP_LIST_FOLD_RIGHT_919(o, s, l) BOOST_PP_LIST_FOLD_LEFT_919(o, s, BOOST_PP_LIST_REVERSE_D(919, l))
# define BOOST_PP_LIST_FOLD_RIGHT_920(o, s, l) BOOST_PP_LIST_FOLD_LEFT_920(o, s, BOOST_PP_LIST_REVERSE_D(920, l))
# define BOOST_PP_LIST_FOLD_RIGHT_921(o, s, l) BOOST_PP_LIST_FOLD_LEFT_921(o, s, BOOST_PP_LIST_REVERSE_D(921, l))
# define BOOST_PP_LIST_FOLD_RIGHT_922(o, s, l) BOOST_PP_LIST_FOLD_LEFT_922(o, s, BOOST_PP_LIST_REVERSE_D(922, l))
# define BOOST_PP_LIST_FOLD_RIGHT_923(o, s, l) BOOST_PP_LIST_FOLD_LEFT_923(o, s, BOOST_PP_LIST_REVERSE_D(923, l))
# define BOOST_PP_LIST_FOLD_RIGHT_924(o, s, l) BOOST_PP_LIST_FOLD_LEFT_924(o, s, BOOST_PP_LIST_REVERSE_D(924, l))
# define BOOST_PP_LIST_FOLD_RIGHT_925(o, s, l) BOOST_PP_LIST_FOLD_LEFT_925(o, s, BOOST_PP_LIST_REVERSE_D(925, l))
# define BOOST_PP_LIST_FOLD_RIGHT_926(o, s, l) BOOST_PP_LIST_FOLD_LEFT_926(o, s, BOOST_PP_LIST_REVERSE_D(926, l))
# define BOOST_PP_LIST_FOLD_RIGHT_927(o, s, l) BOOST_PP_LIST_FOLD_LEFT_927(o, s, BOOST_PP_LIST_REVERSE_D(927, l))
# define BOOST_PP_LIST_FOLD_RIGHT_928(o, s, l) BOOST_PP_LIST_FOLD_LEFT_928(o, s, BOOST_PP_LIST_REVERSE_D(928, l))
# define BOOST_PP_LIST_FOLD_RIGHT_929(o, s, l) BOOST_PP_LIST_FOLD_LEFT_929(o, s, BOOST_PP_LIST_REVERSE_D(929, l))
# define BOOST_PP_LIST_FOLD_RIGHT_930(o, s, l) BOOST_PP_LIST_FOLD_LEFT_930(o, s, BOOST_PP_LIST_REVERSE_D(930, l))
# define BOOST_PP_LIST_FOLD_RIGHT_931(o, s, l) BOOST_PP_LIST_FOLD_LEFT_931(o, s, BOOST_PP_LIST_REVERSE_D(931, l))
# define BOOST_PP_LIST_FOLD_RIGHT_932(o, s, l) BOOST_PP_LIST_FOLD_LEFT_932(o, s, BOOST_PP_LIST_REVERSE_D(932, l))
# define BOOST_PP_LIST_FOLD_RIGHT_933(o, s, l) BOOST_PP_LIST_FOLD_LEFT_933(o, s, BOOST_PP_LIST_REVERSE_D(933, l))
# define BOOST_PP_LIST_FOLD_RIGHT_934(o, s, l) BOOST_PP_LIST_FOLD_LEFT_934(o, s, BOOST_PP_LIST_REVERSE_D(934, l))
# define BOOST_PP_LIST_FOLD_RIGHT_935(o, s, l) BOOST_PP_LIST_FOLD_LEFT_935(o, s, BOOST_PP_LIST_REVERSE_D(935, l))
# define BOOST_PP_LIST_FOLD_RIGHT_936(o, s, l) BOOST_PP_LIST_FOLD_LEFT_936(o, s, BOOST_PP_LIST_REVERSE_D(936, l))
# define BOOST_PP_LIST_FOLD_RIGHT_937(o, s, l) BOOST_PP_LIST_FOLD_LEFT_937(o, s, BOOST_PP_LIST_REVERSE_D(937, l))
# define BOOST_PP_LIST_FOLD_RIGHT_938(o, s, l) BOOST_PP_LIST_FOLD_LEFT_938(o, s, BOOST_PP_LIST_REVERSE_D(938, l))
# define BOOST_PP_LIST_FOLD_RIGHT_939(o, s, l) BOOST_PP_LIST_FOLD_LEFT_939(o, s, BOOST_PP_LIST_REVERSE_D(939, l))
# define BOOST_PP_LIST_FOLD_RIGHT_940(o, s, l) BOOST_PP_LIST_FOLD_LEFT_940(o, s, BOOST_PP_LIST_REVERSE_D(940, l))
# define BOOST_PP_LIST_FOLD_RIGHT_941(o, s, l) BOOST_PP_LIST_FOLD_LEFT_941(o, s, BOOST_PP_LIST_REVERSE_D(941, l))
# define BOOST_PP_LIST_FOLD_RIGHT_942(o, s, l) BOOST_PP_LIST_FOLD_LEFT_942(o, s, BOOST_PP_LIST_REVERSE_D(942, l))
# define BOOST_PP_LIST_FOLD_RIGHT_943(o, s, l) BOOST_PP_LIST_FOLD_LEFT_943(o, s, BOOST_PP_LIST_REVERSE_D(943, l))
# define BOOST_PP_LIST_FOLD_RIGHT_944(o, s, l) BOOST_PP_LIST_FOLD_LEFT_944(o, s, BOOST_PP_LIST_REVERSE_D(944, l))
# define BOOST_PP_LIST_FOLD_RIGHT_945(o, s, l) BOOST_PP_LIST_FOLD_LEFT_945(o, s, BOOST_PP_LIST_REVERSE_D(945, l))
# define BOOST_PP_LIST_FOLD_RIGHT_946(o, s, l) BOOST_PP_LIST_FOLD_LEFT_946(o, s, BOOST_PP_LIST_REVERSE_D(946, l))
# define BOOST_PP_LIST_FOLD_RIGHT_947(o, s, l) BOOST_PP_LIST_FOLD_LEFT_947(o, s, BOOST_PP_LIST_REVERSE_D(947, l))
# define BOOST_PP_LIST_FOLD_RIGHT_948(o, s, l) BOOST_PP_LIST_FOLD_LEFT_948(o, s, BOOST_PP_LIST_REVERSE_D(948, l))
# define BOOST_PP_LIST_FOLD_RIGHT_949(o, s, l) BOOST_PP_LIST_FOLD_LEFT_949(o, s, BOOST_PP_LIST_REVERSE_D(949, l))
# define BOOST_PP_LIST_FOLD_RIGHT_950(o, s, l) BOOST_PP_LIST_FOLD_LEFT_950(o, s, BOOST_PP_LIST_REVERSE_D(950, l))
# define BOOST_PP_LIST_FOLD_RIGHT_951(o, s, l) BOOST_PP_LIST_FOLD_LEFT_951(o, s, BOOST_PP_LIST_REVERSE_D(951, l))
# define BOOST_PP_LIST_FOLD_RIGHT_952(o, s, l) BOOST_PP_LIST_FOLD_LEFT_952(o, s, BOOST_PP_LIST_REVERSE_D(952, l))
# define BOOST_PP_LIST_FOLD_RIGHT_953(o, s, l) BOOST_PP_LIST_FOLD_LEFT_953(o, s, BOOST_PP_LIST_REVERSE_D(953, l))
# define BOOST_PP_LIST_FOLD_RIGHT_954(o, s, l) BOOST_PP_LIST_FOLD_LEFT_954(o, s, BOOST_PP_LIST_REVERSE_D(954, l))
# define BOOST_PP_LIST_FOLD_RIGHT_955(o, s, l) BOOST_PP_LIST_FOLD_LEFT_955(o, s, BOOST_PP_LIST_REVERSE_D(955, l))
# define BOOST_PP_LIST_FOLD_RIGHT_956(o, s, l) BOOST_PP_LIST_FOLD_LEFT_956(o, s, BOOST_PP_LIST_REVERSE_D(956, l))
# define BOOST_PP_LIST_FOLD_RIGHT_957(o, s, l) BOOST_PP_LIST_FOLD_LEFT_957(o, s, BOOST_PP_LIST_REVERSE_D(957, l))
# define BOOST_PP_LIST_FOLD_RIGHT_958(o, s, l) BOOST_PP_LIST_FOLD_LEFT_958(o, s, BOOST_PP_LIST_REVERSE_D(958, l))
# define BOOST_PP_LIST_FOLD_RIGHT_959(o, s, l) BOOST_PP_LIST_FOLD_LEFT_959(o, s, BOOST_PP_LIST_REVERSE_D(959, l))
# define BOOST_PP_LIST_FOLD_RIGHT_960(o, s, l) BOOST_PP_LIST_FOLD_LEFT_960(o, s, BOOST_PP_LIST_REVERSE_D(960, l))
# define BOOST_PP_LIST_FOLD_RIGHT_961(o, s, l) BOOST_PP_LIST_FOLD_LEFT_961(o, s, BOOST_PP_LIST_REVERSE_D(961, l))
# define BOOST_PP_LIST_FOLD_RIGHT_962(o, s, l) BOOST_PP_LIST_FOLD_LEFT_962(o, s, BOOST_PP_LIST_REVERSE_D(962, l))
# define BOOST_PP_LIST_FOLD_RIGHT_963(o, s, l) BOOST_PP_LIST_FOLD_LEFT_963(o, s, BOOST_PP_LIST_REVERSE_D(963, l))
# define BOOST_PP_LIST_FOLD_RIGHT_964(o, s, l) BOOST_PP_LIST_FOLD_LEFT_964(o, s, BOOST_PP_LIST_REVERSE_D(964, l))
# define BOOST_PP_LIST_FOLD_RIGHT_965(o, s, l) BOOST_PP_LIST_FOLD_LEFT_965(o, s, BOOST_PP_LIST_REVERSE_D(965, l))
# define BOOST_PP_LIST_FOLD_RIGHT_966(o, s, l) BOOST_PP_LIST_FOLD_LEFT_966(o, s, BOOST_PP_LIST_REVERSE_D(966, l))
# define BOOST_PP_LIST_FOLD_RIGHT_967(o, s, l) BOOST_PP_LIST_FOLD_LEFT_967(o, s, BOOST_PP_LIST_REVERSE_D(967, l))
# define BOOST_PP_LIST_FOLD_RIGHT_968(o, s, l) BOOST_PP_LIST_FOLD_LEFT_968(o, s, BOOST_PP_LIST_REVERSE_D(968, l))
# define BOOST_PP_LIST_FOLD_RIGHT_969(o, s, l) BOOST_PP_LIST_FOLD_LEFT_969(o, s, BOOST_PP_LIST_REVERSE_D(969, l))
# define BOOST_PP_LIST_FOLD_RIGHT_970(o, s, l) BOOST_PP_LIST_FOLD_LEFT_970(o, s, BOOST_PP_LIST_REVERSE_D(970, l))
# define BOOST_PP_LIST_FOLD_RIGHT_971(o, s, l) BOOST_PP_LIST_FOLD_LEFT_971(o, s, BOOST_PP_LIST_REVERSE_D(971, l))
# define BOOST_PP_LIST_FOLD_RIGHT_972(o, s, l) BOOST_PP_LIST_FOLD_LEFT_972(o, s, BOOST_PP_LIST_REVERSE_D(972, l))
# define BOOST_PP_LIST_FOLD_RIGHT_973(o, s, l) BOOST_PP_LIST_FOLD_LEFT_973(o, s, BOOST_PP_LIST_REVERSE_D(973, l))
# define BOOST_PP_LIST_FOLD_RIGHT_974(o, s, l) BOOST_PP_LIST_FOLD_LEFT_974(o, s, BOOST_PP_LIST_REVERSE_D(974, l))
# define BOOST_PP_LIST_FOLD_RIGHT_975(o, s, l) BOOST_PP_LIST_FOLD_LEFT_975(o, s, BOOST_PP_LIST_REVERSE_D(975, l))
# define BOOST_PP_LIST_FOLD_RIGHT_976(o, s, l) BOOST_PP_LIST_FOLD_LEFT_976(o, s, BOOST_PP_LIST_REVERSE_D(976, l))
# define BOOST_PP_LIST_FOLD_RIGHT_977(o, s, l) BOOST_PP_LIST_FOLD_LEFT_977(o, s, BOOST_PP_LIST_REVERSE_D(977, l))
# define BOOST_PP_LIST_FOLD_RIGHT_978(o, s, l) BOOST_PP_LIST_FOLD_LEFT_978(o, s, BOOST_PP_LIST_REVERSE_D(978, l))
# define BOOST_PP_LIST_FOLD_RIGHT_979(o, s, l) BOOST_PP_LIST_FOLD_LEFT_979(o, s, BOOST_PP_LIST_REVERSE_D(979, l))
# define BOOST_PP_LIST_FOLD_RIGHT_980(o, s, l) BOOST_PP_LIST_FOLD_LEFT_980(o, s, BOOST_PP_LIST_REVERSE_D(980, l))
# define BOOST_PP_LIST_FOLD_RIGHT_981(o, s, l) BOOST_PP_LIST_FOLD_LEFT_981(o, s, BOOST_PP_LIST_REVERSE_D(981, l))
# define BOOST_PP_LIST_FOLD_RIGHT_982(o, s, l) BOOST_PP_LIST_FOLD_LEFT_982(o, s, BOOST_PP_LIST_REVERSE_D(982, l))
# define BOOST_PP_LIST_FOLD_RIGHT_983(o, s, l) BOOST_PP_LIST_FOLD_LEFT_983(o, s, BOOST_PP_LIST_REVERSE_D(983, l))
# define BOOST_PP_LIST_FOLD_RIGHT_984(o, s, l) BOOST_PP_LIST_FOLD_LEFT_984(o, s, BOOST_PP_LIST_REVERSE_D(984, l))
# define BOOST_PP_LIST_FOLD_RIGHT_985(o, s, l) BOOST_PP_LIST_FOLD_LEFT_985(o, s, BOOST_PP_LIST_REVERSE_D(985, l))
# define BOOST_PP_LIST_FOLD_RIGHT_986(o, s, l) BOOST_PP_LIST_FOLD_LEFT_986(o, s, BOOST_PP_LIST_REVERSE_D(986, l))
# define BOOST_PP_LIST_FOLD_RIGHT_987(o, s, l) BOOST_PP_LIST_FOLD_LEFT_987(o, s, BOOST_PP_LIST_REVERSE_D(987, l))
# define BOOST_PP_LIST_FOLD_RIGHT_988(o, s, l) BOOST_PP_LIST_FOLD_LEFT_988(o, s, BOOST_PP_LIST_REVERSE_D(988, l))
# define BOOST_PP_LIST_FOLD_RIGHT_989(o, s, l) BOOST_PP_LIST_FOLD_LEFT_989(o, s, BOOST_PP_LIST_REVERSE_D(989, l))
# define BOOST_PP_LIST_FOLD_RIGHT_990(o, s, l) BOOST_PP_LIST_FOLD_LEFT_990(o, s, BOOST_PP_LIST_REVERSE_D(990, l))
# define BOOST_PP_LIST_FOLD_RIGHT_991(o, s, l) BOOST_PP_LIST_FOLD_LEFT_991(o, s, BOOST_PP_LIST_REVERSE_D(991, l))
# define BOOST_PP_LIST_FOLD_RIGHT_992(o, s, l) BOOST_PP_LIST_FOLD_LEFT_992(o, s, BOOST_PP_LIST_REVERSE_D(992, l))
# define BOOST_PP_LIST_FOLD_RIGHT_993(o, s, l) BOOST_PP_LIST_FOLD_LEFT_993(o, s, BOOST_PP_LIST_REVERSE_D(993, l))
# define BOOST_PP_LIST_FOLD_RIGHT_994(o, s, l) BOOST_PP_LIST_FOLD_LEFT_994(o, s, BOOST_PP_LIST_REVERSE_D(994, l))
# define BOOST_PP_LIST_FOLD_RIGHT_995(o, s, l) BOOST_PP_LIST_FOLD_LEFT_995(o, s, BOOST_PP_LIST_REVERSE_D(995, l))
# define BOOST_PP_LIST_FOLD_RIGHT_996(o, s, l) BOOST_PP_LIST_FOLD_LEFT_996(o, s, BOOST_PP_LIST_REVERSE_D(996, l))
# define BOOST_PP_LIST_FOLD_RIGHT_997(o, s, l) BOOST_PP_LIST_FOLD_LEFT_997(o, s, BOOST_PP_LIST_REVERSE_D(997, l))
# define BOOST_PP_LIST_FOLD_RIGHT_998(o, s, l) BOOST_PP_LIST_FOLD_LEFT_998(o, s, BOOST_PP_LIST_REVERSE_D(998, l))
# define BOOST_PP_LIST_FOLD_RIGHT_999(o, s, l) BOOST_PP_LIST_FOLD_LEFT_999(o, s, BOOST_PP_LIST_REVERSE_D(999, l))
# define BOOST_PP_LIST_FOLD_RIGHT_1000(o, s, l) BOOST_PP_LIST_FOLD_LEFT_1000(o, s, BOOST_PP_LIST_REVERSE_D(1000, l))
# define BOOST_PP_LIST_FOLD_RIGHT_1001(o, s, l) BOOST_PP_LIST_FOLD_LEFT_1001(o, s, BOOST_PP_LIST_REVERSE_D(1001, l))
# define BOOST_PP_LIST_FOLD_RIGHT_1002(o, s, l) BOOST_PP_LIST_FOLD_LEFT_1002(o, s, BOOST_PP_LIST_REVERSE_D(1002, l))
# define BOOST_PP_LIST_FOLD_RIGHT_1003(o, s, l) BOOST_PP_LIST_FOLD_LEFT_1003(o, s, BOOST_PP_LIST_REVERSE_D(1003, l))
# define BOOST_PP_LIST_FOLD_RIGHT_1004(o, s, l) BOOST_PP_LIST_FOLD_LEFT_1004(o, s, BOOST_PP_LIST_REVERSE_D(1004, l))
# define BOOST_PP_LIST_FOLD_RIGHT_1005(o, s, l) BOOST_PP_LIST_FOLD_LEFT_1005(o, s, BOOST_PP_LIST_REVERSE_D(1005, l))
# define BOOST_PP_LIST_FOLD_RIGHT_1006(o, s, l) BOOST_PP_LIST_FOLD_LEFT_1006(o, s, BOOST_PP_LIST_REVERSE_D(1006, l))
# define BOOST_PP_LIST_FOLD_RIGHT_1007(o, s, l) BOOST_PP_LIST_FOLD_LEFT_1007(o, s, BOOST_PP_LIST_REVERSE_D(1007, l))
# define BOOST_PP_LIST_FOLD_RIGHT_1008(o, s, l) BOOST_PP_LIST_FOLD_LEFT_1008(o, s, BOOST_PP_LIST_REVERSE_D(1008, l))
# define BOOST_PP_LIST_FOLD_RIGHT_1009(o, s, l) BOOST_PP_LIST_FOLD_LEFT_1009(o, s, BOOST_PP_LIST_REVERSE_D(1009, l))
# define BOOST_PP_LIST_FOLD_RIGHT_1010(o, s, l) BOOST_PP_LIST_FOLD_LEFT_1010(o, s, BOOST_PP_LIST_REVERSE_D(1010, l))
# define BOOST_PP_LIST_FOLD_RIGHT_1011(o, s, l) BOOST_PP_LIST_FOLD_LEFT_1011(o, s, BOOST_PP_LIST_REVERSE_D(1011, l))
# define BOOST_PP_LIST_FOLD_RIGHT_1012(o, s, l) BOOST_PP_LIST_FOLD_LEFT_1012(o, s, BOOST_PP_LIST_REVERSE_D(1012, l))
# define BOOST_PP_LIST_FOLD_RIGHT_1013(o, s, l) BOOST_PP_LIST_FOLD_LEFT_1013(o, s, BOOST_PP_LIST_REVERSE_D(1013, l))
# define BOOST_PP_LIST_FOLD_RIGHT_1014(o, s, l) BOOST_PP_LIST_FOLD_LEFT_1014(o, s, BOOST_PP_LIST_REVERSE_D(1014, l))
# define BOOST_PP_LIST_FOLD_RIGHT_1015(o, s, l) BOOST_PP_LIST_FOLD_LEFT_1015(o, s, BOOST_PP_LIST_REVERSE_D(1015, l))
# define BOOST_PP_LIST_FOLD_RIGHT_1016(o, s, l) BOOST_PP_LIST_FOLD_LEFT_1016(o, s, BOOST_PP_LIST_REVERSE_D(1016, l))
# define BOOST_PP_LIST_FOLD_RIGHT_1017(o, s, l) BOOST_PP_LIST_FOLD_LEFT_1017(o, s, BOOST_PP_LIST_REVERSE_D(1017, l))
# define BOOST_PP_LIST_FOLD_RIGHT_1018(o, s, l) BOOST_PP_LIST_FOLD_LEFT_1018(o, s, BOOST_PP_LIST_REVERSE_D(1018, l))
# define BOOST_PP_LIST_FOLD_RIGHT_1019(o, s, l) BOOST_PP_LIST_FOLD_LEFT_1019(o, s, BOOST_PP_LIST_REVERSE_D(1019, l))
# define BOOST_PP_LIST_FOLD_RIGHT_1020(o, s, l) BOOST_PP_LIST_FOLD_LEFT_1020(o, s, BOOST_PP_LIST_REVERSE_D(1020, l))
# define BOOST_PP_LIST_FOLD_RIGHT_1021(o, s, l) BOOST_PP_LIST_FOLD_LEFT_1021(o, s, BOOST_PP_LIST_REVERSE_D(1021, l))
# define BOOST_PP_LIST_FOLD_RIGHT_1022(o, s, l) BOOST_PP_LIST_FOLD_LEFT_1022(o, s, BOOST_PP_LIST_REVERSE_D(1022, l))
# define BOOST_PP_LIST_FOLD_RIGHT_1023(o, s, l) BOOST_PP_LIST_FOLD_LEFT_1023(o, s, BOOST_PP_LIST_REVERSE_D(1023, l))
# define BOOST_PP_LIST_FOLD_RIGHT_1024(o, s, l) BOOST_PP_LIST_FOLD_LEFT_1024(o, s, BOOST_PP_LIST_REVERSE_D(1024, l))
#
# endif
