autogen definitions test_endswap.tpl;

int_type = {
	name		= short ;
	value		= '0x3210' ;
	format		= FMT_SHORT ;
	} ;

int_type = {
	name		= int ;
	value		= '0x76543210' ;
	format		= FMT_INT ;
	} ;

int_type = {
	name		= int64_t ;
	value		= '0x0807050540302010LL' ;
	format		= FMT_INT64 ;
	} ;

int_size = {
	name		= 16 ;
	typename	= int16_t ;
	value		= '0x4142' ;
	strval		= "AB" ;
	} ;

int_size = {
	name		= 32 ;
	typename	= int32_t ;
	value		= '0x30313233' ;
	strval		= "0123" ;
	} ;

int_size = {
	name		= 64 ;
	typename	= int64_t ;
	value		= '0x3031323334353637' ;
	strval		= "01234567" ;
	} ;
