/*
Copyright Rene Rivera 2015-2016
Distributed under the Boost Software License, Version 1.0.
(See accompanying file LICENSE_1_0.txt or copy at
http://www.boost.org/LICENSE_1_0.txt)
*/

#ifndef BOOST_PREDEF_VERSION_H
#define BOOST_PREDEF_VERSION_H

#include <boost/predef/version_number.h>

#define BOOST_PREDEF_VERSION BOOST_VERSION_NUMBER(1,15,1)

#endif
