/**
 * This file has no copyright assigned and is placed in the Public Domain.
 * This file is part of the PortAudio library.
 */
#ifndef _INC_PROPKEY_PA
#define _INC_PROPKEY_PA

#ifndef DEFINE_API_PKEY
#define DEFINE_API_PKEY(name, managed_name, l, w1, w2, b1, b2, b3, b4, b5, b6, b7, b8, pid) \
    DEFINE_PROPERTYKEY(name, l, w1, w2, b1, b2, b3, b4, b5, b6, b7, b8, pid)
#endif

#endif /* _INC_PROPKEY_PA */
